/*
 * Copyright 2008, The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <pthread.h>
#include <dlfcn.h>

#include "hardware_legacy/wifi.h"
#include "libwpa_client/wpa_ctrl.h"

#define LOG_TAG "WifiHW"
#include "cutils/log.h"
#include "cutils/memory.h"
#include "cutils/misc.h"
#include "cutils/properties.h"
#include "private/android_filesystem_config.h"
#include "wlan_adp_intf.h"

//added by derek
#define WIFI_DEBUG

#ifdef WIFI_DEBUG 
#define ENTER() LOGD("enter into %s", __FUNCTION__);
#define LEAVE() LOGD("leave from %s", __FUNCTION__);
#else
#define ENTER() do { } while (0)
#define LEAVE() do { } while (0)
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

static struct wpa_ctrl *ctrl_conn;
static struct wpa_ctrl *monitor_conn;
//this global static varible describes if we need reset socket or not.
static int need_reset_sock = FALSE;

extern int do_dhcp();
extern int ifc_init();
extern void ifc_close();
extern char *dhcp_lasterror();
extern void get_dhcp_info();

static char iface[PROPERTY_VALUE_MAX];
// TODO: use new ANDROID_SOCKET mechanism, once support for multiple
// sockets is in

#define WIFI_TEST_INTERFACE		"sta"
static const char IFACE_DIR[]           = "/data/misc/wifi/sockets/server";
static const char FIRMWARE_LOADER[]     = "wlan_loader";
static const char SUPPLICANT_NAME[]     = "wpa_supplicant";
static const char SUPP_PROP_NAME[]      = "init.svc.wpa_supplicant";
static const char SUPP_CONFIG_TEMPLATE[]= "/system/etc/wifi/wpa_supplicant.conf";
static const char SUPP_CONFIG_FILE[]    = "/data/misc/wifi/wpa_supplicant.conf";
//static const char SUPP_LOG_FILE[]    = "/local/log/wpa_log.txt";
//static const char SUPP_LOG_PRE_FILE[]    = "/local/log/wpa_log_pre.txt";
//static const char SUPP_CLI_IFACE_DIR[]    = "/data/misc/wifi/sockets/clients";
//static const char DRIVER_REF_PROP_NAME[] = "wlan.driver.ref";

const WLAN_ADP_LIB_Functions *(*gWlanAdpInit)(void);
const WLAN_ADP_LIB_Functions *gWlanAdpFuncs = NULL;

int do_dhcp_request(int *ipaddr, int *gateway, int *mask,
                    int *dns1, int *dns2, int *server, int *lease) {
    ENTER();
    /* For test driver, always report success */
    if (strcmp(iface, WIFI_TEST_INTERFACE) == 0)
        return 0;

    if (ifc_init() < 0)
        return -1;

    if (do_dhcp(iface) < 0) {
        ifc_close();
        return -1;
    }
    ifc_close();
    get_dhcp_info(ipaddr, gateway, mask, dns1, dns2, server, lease);
    LEAVE();

    return 0;
}

const char *get_dhcp_error_string() {
    return dhcp_lasterror();
}

static void *load_dynamic_lib(char * libname)
{
    void * dlHandle = 0;

    dlHandle = dlopen(libname, RTLD_NOW);
    if (dlHandle == NULL) {
        LOGE("load wlan shared library %s error", libname);
        return 0;
    }
    LOGD("load wlan shared library %s successfully", libname);

    return dlHandle;
}


int wifi_load_driver()
{
    void * dlHandle = 0;
    char chipname[32];
    char libname[32];

    if (gWlanAdpFuncs == NULL) {
        LOGD("in %s, gWlanAdpFuncs is NULL", __FUNCTION__);

        memset(chipname, 0, 32);
        memset(libname, 0, 32);
        property_get("hw.wifi.chipset", chipname, NULL);
        sprintf(libname, "libwlan%s.so", chipname);

        dlHandle = load_dynamic_lib(libname);
        if (dlHandle == 0) {
            return -1;
        }

        gWlanAdpInit = (const WLAN_ADP_LIB_Functions *(*)(void))dlsym(dlHandle, "WLAN_ADP_LIB_Init");
        if (gWlanAdpInit == NULL) {
            LOGE("gWlanAdpInit is NULL");
            return -1;
        } else {
            gWlanAdpFuncs = gWlanAdpInit();
        }
    }

    if (gWlanAdpFuncs == NULL) {
        LOGE("gWlanAdpFuncs is still NULL");
        return -1;
    }
    //check if vendor has implemented this function callback.
    if (gWlanAdpFuncs->wlan_load_driver == 0) {
        LOGE("wlan_load_driver callback is 0");
        return -1;
    }

    if (gWlanAdpFuncs->wlan_load_driver() < 0) {
        LOGE("call wlan_load_driver, return fail");
        return -1;
    }
    LOGD("call wlan_load_driver, return ok");

    return 0;
}

int wifi_unload_driver()
{
    void * dlHandle = 0;
    char chipname[32];
    char libname[32];

    if (gWlanAdpFuncs == NULL) {
        LOGD("in %s, gWlanAdpFuncs is NULL", __FUNCTION__);

        memset(chipname, 0, 32);
        memset(libname, 0, 32);
        property_get("hw.wifi.chipset", chipname, NULL);
        sprintf(libname, "libwlan%s.so", chipname);

        dlHandle = load_dynamic_lib(libname);
        if (dlHandle == 0) {
            return -1;
        }

        gWlanAdpInit = (const WLAN_ADP_LIB_Functions *(*)(void))dlsym(dlHandle, "WLAN_ADP_LIB_Init");
        if (gWlanAdpInit == NULL) {
            LOGE("gWlanAdpInit is NULL");
            return -1;
        } else {
            gWlanAdpFuncs = gWlanAdpInit();
        }
    }

    if (gWlanAdpFuncs == NULL) {
        LOGE("gWlanAdpFuncs is still NULL");
        return -1;
    }

    //check if vendor has implemented this function callback.
    if (gWlanAdpFuncs->wlan_unload_driver == 0) {
        LOGE("wlan_unload_driver callback is 0");
        return -1;
    }

    if (gWlanAdpFuncs->wlan_unload_driver() < 0) {
        LOGE("call wlan_unload_driver, return fail");
        return -1;
    }
    LOGD("call wlan_unload_driver, return ok");

    return 0;
}

static int control_supplicant(int startIt)
{
    char supp_status[PROPERTY_VALUE_MAX] = {'\0'};
    const char *ctrl_prop = (startIt ? "ctl.start" : "ctl.stop");
    const char *desired_status = (startIt ? "running" : "stopped");
    int count = 20; /* wait at most 20 seconds for completion */

    ENTER();
    if (property_get(SUPP_PROP_NAME, supp_status, NULL)
        && strcmp(supp_status, desired_status) == 0) {
        LOGD("in control_supplicant, supplicant is already running\n");
        return 0;  /* supplicant already running */
    }
    LOGD("in control_supplicant, supplicant will be launched\n");
    property_set(ctrl_prop, SUPPLICANT_NAME);

    while (count-- > 0) {
        usleep(1000000);
        if (property_get(SUPP_PROP_NAME, supp_status, NULL)) {
            if (strcmp(supp_status, desired_status) == 0)
                return 0;
        }
    }
    LEAVE();

    return -1;
}

int ensure_config_file_exists()
{
    char buf[2048];
    int srcfd, destfd;
    int nread;

    ENTER();
    if (access(SUPP_CONFIG_FILE, R_OK|W_OK) == 0) {
        return 0;
    } else if (errno != ENOENT) {
        LOGE("Cannot access \"%s\": %s", SUPP_CONFIG_FILE, strerror(errno));
        return -1;
    }

    srcfd = open(SUPP_CONFIG_TEMPLATE, O_RDONLY);
    if (srcfd < 0) {
        LOGE("Cannot open \"%s\": %s", SUPP_CONFIG_TEMPLATE, strerror(errno));
        return -1;
    }

    destfd = open(SUPP_CONFIG_FILE, O_CREAT|O_WRONLY, 0660);
    if (destfd < 0) {
        close(srcfd);
        LOGE("Cannot create \"%s\": %s", SUPP_CONFIG_FILE, strerror(errno));
        return -1;
    }

    while ((nread = read(srcfd, buf, sizeof(buf))) != 0) {
        if (nread < 0) {
            LOGE("Error reading \"%s\": %s", SUPP_CONFIG_TEMPLATE, strerror(errno));
            close(srcfd);
            close(destfd);
            unlink(SUPP_CONFIG_FILE);
            return -1;
        }
        write(destfd, buf, nread);
    }

    close(destfd);
    close(srcfd);

    if (chown(SUPP_CONFIG_FILE, AID_SYSTEM, AID_WIFI) < 0) {
        LOGE("Error changing group ownership of %s to %d: %s",
             SUPP_CONFIG_FILE, AID_WIFI, strerror(errno));
        unlink(SUPP_CONFIG_FILE);
        return -1;
    }
    LEAVE();

    return 0;
}

//deleted by derek in OMS1.6
#if 0
static void prepare_for_supplicant()
{
    DIR * dp = NULL;
    struct dirent * dirp = NULL;
    char iface_name[128];

    ENTER();
    //delete wpa_log_pre.txt file
    if (access(SUPP_LOG_PRE_FILE, F_OK|W_OK) == 0) {
        if (unlink(SUPP_LOG_PRE_FILE) == 0)
            LOGD("delete %s successfully", SUPP_LOG_PRE_FILE);
        else
            LOGD("delete %s unsuccessfully", SUPP_LOG_PRE_FILE);
    }

    //rename wpa_log.txt file
    if (access(SUPP_LOG_FILE, F_OK|W_OK) == 0) {
        if (rename(SUPP_LOG_FILE, SUPP_LOG_PRE_FILE) == 0)
            LOGD("rename %s successfully", SUPP_LOG_FILE);
        else
            LOGD("rename %s unsuccessfully", SUPP_LOG_FILE);
    }

    //delete all socket local file in /data/misc/wifi/sockets
    if ((dp = opendir(SUPP_CLI_IFACE_DIR)) == NULL) {
        LOGD("open dir %s unsuccessfully\n", SUPP_CLI_IFACE_DIR);
        return;
    }

    while ((dirp = readdir(dp)) != NULL)
    {
        if(strncmp(dirp->d_name, "wpa_ctrl_", 9) == NULL)
        {
            sprintf(iface_name,"%s/%s", SUPP_CLI_IFACE_DIR, dirp->d_name);
            if (unlink(iface_name) == 0)
                LOGD("delete %s successfully\n", dirp->d_name);
            else
                LOGD("delete %s unsuccessfully\n", dirp->d_name);
        }
    }
    closedir(dp);
 
    LEAVE();
    return;
}
#endif

int wifi_start_supplicant()
{
    ENTER();
    /* Before starting the daemon, make sure its config file exists */
    if (ensure_config_file_exists() < 0) {
        LOGE("Wi-Fi will not be enabled");
        return -1;
    }
    //prepare_for_supplicant();

    return control_supplicant(1);
}

int wifi_stop_supplicant()
{
    ENTER();

    return control_supplicant(0);
}

int wifi_connect_to_supplicant()
{
    char ifname[256];
    char supp_status[PROPERTY_VALUE_MAX] = {'\0'};
    static int cleaned_up = 0;

    ENTER();
    /* Make sure supplicant is running */
    if (!property_get(SUPP_PROP_NAME, supp_status, NULL)
            || strcmp(supp_status, "running") != 0) {
        LOGE("Supplicant not running, cannot connect");
        return -1;
    }

    property_get("wifi.interface", iface, WIFI_TEST_INTERFACE);

    if (access(IFACE_DIR, F_OK) == 0) {
        snprintf(ifname, sizeof(ifname), "%s/%s", IFACE_DIR, iface);
    } else {
        strlcpy(ifname, iface, sizeof(ifname));
    }
    LOGD("in wifi_connect_to_supplicant, ifname is %s", ifname);

    ctrl_conn = wpa_ctrl_open(ifname);
    if (ctrl_conn == NULL) {
        LOGD("Unable to open connection to supplicant on \"%s\": %s",
             ifname, strerror(errno));
        /*
         * errno == ENOENT means the supplicant daemon isn't
         * running. Take this opportunity to clear out any
         * stale socket files that might be left over. Note
         * there's a possible race with the command line client
         * trying to connect to the daemon, but it would require
         * that the supplicant be started and the command line
         * client connect to it during the window between the
         * error check and the removal of the files. And in
         * any event, the remedy is that the user would simply
         * have to run the command line program again.
         */
        if (!cleaned_up && (errno == ENOENT || errno == EADDRINUSE)) {
            cleaned_up = 1; /* do this just once */
            wpa_ctrl_cleanup();
        }
        return -1;
    }
    monitor_conn = wpa_ctrl_open(ifname);
    if (monitor_conn == NULL) {
        wpa_ctrl_close(ctrl_conn);
        ctrl_conn = NULL;
        return -1;
    }
    if (wpa_ctrl_attach(monitor_conn) != 0) {
        wpa_ctrl_close(monitor_conn);
        wpa_ctrl_close(ctrl_conn);
        ctrl_conn = monitor_conn = NULL;
        return -1;
    }
    LEAVE();

    return 0;
}

int wifi_send_command(struct wpa_ctrl *ctrl, const char *cmd, char *reply, size_t *reply_len)
{
    int ret;
    static int try_num = 0;
#define WIFI_SEND_FAIL_TRY_MAX_NUM 5

    if (ctrl_conn == NULL || need_reset_sock == TRUE) {
        LOGV("Not connected to wpa_supplicant - \"%s\" command dropped.\n", cmd);
        return -1;
    }
    ret = wpa_ctrl_request(ctrl, cmd, strlen(cmd), reply, reply_len, NULL);
    if (ret == -3) {
        if (++ try_num >= WIFI_SEND_FAIL_TRY_MAX_NUM) {
        	//set need_reset_sock to TRUE to notify event thread to process for it.
        	LOGD("send fails %d times, set need_reset_sock to TRUE", WIFI_SEND_FAIL_TRY_MAX_NUM);
    		need_reset_sock = TRUE;
    		try_num = 0;
        }
        return -1;
    } else if (ret == -2) {
        LOGD("'%s' command timed out.\n", cmd);
        return -2;
    } else if (ret < 0 || strncmp(reply, "FAIL", 4) == 0) {
        return -1;
    }

    //reset try_num to zero if it sends successfully.
    try_num = 0;
    if (strncmp(cmd, "PING", 4) == 0) {
        reply[*reply_len] = '\0';
    }

    return 0;
}

int wifi_wait_for_event(char *buf, size_t buflen)
{
    size_t nread = buflen - 1;
    int fd;
    fd_set rfds;
    int result;
    struct timeval tval;
    struct timeval *tptr;
    
    //ENTER();
    if (monitor_conn == NULL) {
        LOGD("monitor_conn is NULL");
        sleep(1);
        return 0;
    }

    while (1) {
    	if (need_reset_sock == TRUE) {
        	LOGD("try to reset local socket by default for wifi feature robustness");
	        wifi_close_supplicant_connection();
        	wifi_connect_to_supplicant();
        	need_reset_sock = FALSE;
    	}
        
    	if(wpa_ctrl_pending_timeout(monitor_conn, 2, 0) > 0) {
    		//need call recv func
        	result = wpa_ctrl_recv(monitor_conn, buf, &nread);
        	break;
    	}
    }

    if (result < 0) {
        LOGD("wpa_ctrl_recv failed: %s\n", strerror(errno));
        return -1;
    }
    buf[nread] = '\0';
    /* LOGD("wait_for_event: result=%d nread=%d string=\"%s\"\n", result, nread, buf); */
    /* Check for EOF on the socket */
    if (result == 0 && nread == 0) {
        /* Fabricate an event to pass up */
        LOGD("Received EOF on supplicant socket\n");
        strncpy(buf, WPA_EVENT_TERMINATING " - signal 0 received", buflen-1);
        buf[buflen-1] = '\0';
        return strlen(buf);
    }
    /*
     * Events strings are in the format
     *
     *     <N>CTRL-EVENT-XXX 
     *
     * where N is the message level in numerical form (0=VERBOSE, 1=DEBUG,
     * etc.) and XXX is the event name. The level information is not useful
     * to us, so strip it off.
     */
    if (buf[0] == '<') {
        char *match = strchr(buf, '>');
        if (match != NULL) {
            nread -= (match+1-buf);
            memmove(buf, match+1, nread+1);
        }
    }
    //LEAVE();

    return nread;
}

void wifi_close_supplicant_connection()
{
    ENTER();
    if (ctrl_conn != NULL) {
        wpa_ctrl_close(ctrl_conn);
        ctrl_conn = NULL;
    }
    if (monitor_conn != NULL) {
        wpa_ctrl_close(monitor_conn);
        monitor_conn = NULL;
    }
    LEAVE();
}

int wifi_command(const char *command, char *reply, size_t *reply_len)
{
    //ENTER();
    return wifi_send_command(ctrl_conn, command, reply, reply_len);
}
