/*
 * Copyright 2009, Borqs
 *
 * File: wlan_adp_intf.h
 *
 * Revision: 1.0 RC1
 *
 * Author: Derek.hong
 *
 * Date: 2009/4/15
 *
 * Description:
 *     Define wlan adaption interfaces between libhardware library and plugable shared library of vendor.
 *     and also the interfaces between wpa_supplicant and plugable shared library of vendor.
 */

#ifndef _WLAN_ADP_INTF_H_
#define _WLAN_ADP_INTF_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <net/if.h>
#include <sys/types.h>

/*-----------------------------------------------------------------------*/
/*the below data structure and  interfaces are for libhardware.so library*/
/*-----------------------------------------------------------------------*/
/*
 * wlan load driver callback function prototype
 * WLAN vendor shared library must define it,
 *
 * @param void
 * @return int, 0 on success, -1 on failure.
 */
typedef int (*WLAN_ADP_Load_Driver)(void);

/*
 * wlan unload driver callback function prototype
 * WLAN vendor shared library must define it,
 *
 * @param void
 * @return int, 0 on success, -1 on failure.
 */
typedef int (*WLAN_ADP_Unload_Driver)(void);


/*--------------------------------------------------------------*/
/*the below data structure and interfaces are for wpa_supplicant*/
/*--------------------------------------------------------------*/
/**
 * struct wpa_driver_capa - Driver capability information
 */
#ifndef WPA_DRIVER_CAPA
#define WPA_DRIVER_CAPA
struct wpa_driver_capa {
#define WPA_DRIVER_CAPA_KEY_MGMT_WPA        0x00000001
#define WPA_DRIVER_CAPA_KEY_MGMT_WPA2       0x00000002
#define WPA_DRIVER_CAPA_KEY_MGMT_WPA_PSK    0x00000004
#define WPA_DRIVER_CAPA_KEY_MGMT_WPA2_PSK   0x00000008
#define WPA_DRIVER_CAPA_KEY_MGMT_WPA_NONE   0x00000010
    unsigned int key_mgmt;

#define WPA_DRIVER_CAPA_ENC_WEP40   0x00000001
#define WPA_DRIVER_CAPA_ENC_WEP104  0x00000002
#define WPA_DRIVER_CAPA_ENC_TKIP    0x00000004
#define WPA_DRIVER_CAPA_ENC_CCMP    0x00000008
    unsigned int enc;

#define WPA_DRIVER_AUTH_OPEN        0x00000001
#define WPA_DRIVER_AUTH_SHARED      0x00000002
#define WPA_DRIVER_AUTH_LEAP        0x00000004
    unsigned int auth;

/* Driver generated WPA/RSN IE */
#define WPA_DRIVER_FLAGS_DRIVER_IE  0x00000001
#define WPA_DRIVER_FLAGS_SET_KEYS_AFTER_ASSOC 0x00000002
#define WPA_DRIVER_FLAGS_USER_SPACE_MLME 0x00000004
    unsigned int flags;
};
#endif

/* log record function */
typedef void (*wprintf)(int level, char *fmt, ...);

#define v_printf(level, fmt, args...) \
    if (gDrvData != NULL && gDrvData->log != NULL) { \
        gDrvData->log(level, "WPA VENDOR: "fmt, ##args); \
    }

enum { VMSG_MSGDUMP, VMSG_DEBUG, VMSG_INFO, VMSG_WARNING, VMSG_ERROR };

struct wpa_driver_wext_data {
    void *ctx;
    int event_sock;
    int ioctl_sock;
    int mlme_sock;
    char ifname[IFNAMSIZ + 1];
    int ifindex;
    int ifindex2;
    int if_removed;
    uint8_t *assoc_req_ies;
    size_t assoc_req_ies_len;
    uint8_t *assoc_resp_ies;
    size_t assoc_resp_ies_len;
    struct wpa_driver_capa capa;
    int has_capability;
    int we_version_compiled;

    /* for set_auth_alg fallback */
    int use_crypt;
    int auth_alg_fallback;

    int operstate;

    char mlmedev[IFNAMSIZ + 1];

    int scan_complete_events;

    //added by borqs for OMS platform.
    //this pointer points vendor data structure.
    void * vendor_data;
    //this function pointer points to log record function.
    wprintf log;
};

/*
 * wlan vendor driver command callback function prototype
 * WLAN vendor shared library must define it,
 *
 * @param void *priv: struct wpa_driver_wext_data * 
 *        char *cmd: command name
 *        char *buf: buffer of response
 *        size_t buf_len: buf length
 * @return int: return value of the command.
 */
typedef int (*WLAN_ADP_Wpa_Driver_Vendor_Driver_Cmd)(void *priv, char *cmd, char *buf, size_t buf_len);

/*
 * wlan vendor event process callback function prototype
 * WLAN vendor shared library can define it if it need process event,
 *
 * @param void *priv: struct wpa_driver_wext_data * 
 *        char *custom: custom buf 
 * @return int, 0 - on success, -1 on failure.
 */
typedef int (*WLAN_ADP_Wpa_Driver_Vendor_Event_Custom)(void *priv, char *custom); 

/*
 * wlan vendor data initialization callback function prototype
 * WLAN vendor shared library can define it if it need have its own data,
 * this function need allocate memory and initialize it.
 *
 * @param void *priv: struct wpa_driver_wext_data * 
 * @return int, 0 - on success, -1 on failure.
 */
typedef int (*WLAN_ADP_Wpa_Driver_Init_Vendor_Data)(void *priv);

/*
 * wlan vendor data deinitialization callback function prototype
 * WLAN vendor shared library can define it if it has defined init function.,
 * this function need free the memory it has allocated. 
 *
 * @param void *priv: struct wpa_driver_wext_data * 
 * @return int, 0 - on success, -1 on failure.
 */
typedef int (*WLAN_ADP_Wpa_Driver_Deinit_Vendor_Data)(void *priv);

typedef struct
{
    WLAN_ADP_Load_Driver wlan_load_driver;
    WLAN_ADP_Unload_Driver wlan_unload_driver;
    WLAN_ADP_Wpa_Driver_Vendor_Driver_Cmd wpa_driver_vendor_driver_cmd;
    WLAN_ADP_Wpa_Driver_Vendor_Event_Custom wpa_driver_vendor_event_custom;
    WLAN_ADP_Wpa_Driver_Init_Vendor_Data wpa_driver_init_vendor_data;
    WLAN_ADP_Wpa_Driver_Deinit_Vendor_Data wpa_driver_deinit_vendor_data;
} WLAN_ADP_LIB_Functions;

/*
 * WLAN vendor shared library must define WLAN_ADP_LIB_Init,
 * it will return these callbacks.
 *
 * @param void 
 * @return *WLAN_ADP_LIB_Functions
 */
const WLAN_ADP_LIB_Functions *WLAN_ADP_LIB_Init(void);

#ifdef __cplusplus
}
#endif

#endif
