/*
 * Copyright 2009, Borqs
 *
 * File: mrvl8688.c 
 *
 * Revision: 1.0 RC1
 *
 * Author: Derek.hong
 *
 * Date: 2009/4/15
 *
 * Description:
 *     Implement plugable shared library for marvell 8688 chip.
 */

#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <cutils/properties.h>
#include <cutils/misc.h>
#include "wlan_adp_intf.h"

#define LOG_TAG "WifiVdr"
#include <cutils/log.h>

static const char DRIVER_MODULE_NAME[]  = "sd8688";
static const char DRIVER_MODULE_TAG[]   = "sd8688 ";
static const char DRIVER_MODULE_PATH[]  = "/opl/modules/sd8688.ko";
static const char DRIVER_PROP_NAME[]    = "wlan.driver.status";
static const char DRIVER_REF_PROP_NAME[]= "wlan.driver.ref"; //added by li yanbing
static const char MODULE_FILE[]         = "/proc/modules";

#define WLAN_VDR_DEBUG

#ifdef WLAN_VDR_DEBUG 
#define ENTER() LOGD("enter into %s", __FUNCTION__);
#define LEAVE() LOGD("leave from %s", __FUNCTION__);
#else
#define ENTER() do { } while (0)
#define LEAVE() do { } while (0)
#endif

struct wifi_driver_lock{
        pthread_mutex_t lock;
        int ref;
        int ref_num_for_sysserv_reboot;
};

static struct wifi_driver_lock driver_lock={
        PTHREAD_MUTEX_INITIALIZER,
        0,
        0
};

extern int init_module(void *, unsigned long, const char *);
extern int delete_module(const char *, unsigned int);
extern int wpa_driver_vendor_driver_cmd(void *priv, char *cmd, char *buf, size_t buf_len);
extern int wpa_driver_vendor_event_custom(void *priv, char *custom); 
extern int wpa_driver_init_vendor_data(void *priv);
extern int wpa_driver_deinit_vendor_data(void *priv);

static int insmod(const char *filename)
{
    void *module;
    unsigned int size;
    int ret;

    ENTER();
    module = load_file(filename, &size);
    if (!module)
        return -1;

    ret = init_module(module, size, "");

    free(module);
    LEAVE();

    return ret;
}

static int rmmod(const char *modname)
{
    int ret = -1;
    int maxtry = 10;

    ENTER();
    while (maxtry-- > 0) {
        ret = delete_module(modname, O_NONBLOCK | O_EXCL);
        if (ret < 0 && errno == EAGAIN)
            usleep(500000);
        else
            break;
    }

    if (ret != 0)
        LOGD("Unable to unload driver module \"%s\": %s\n",
             modname, strerror(errno));
    LEAVE();

    return ret;
}

static int check_driver_loaded() {
    char driver_ref[PROPERTY_VALUE_MAX];
    FILE *proc;
    char line[sizeof(DRIVER_MODULE_TAG)+10];

    ENTER();
    if (driver_lock.ref == 0 && property_get(DRIVER_REF_PROP_NAME, driver_ref, NULL)){
        if (atoi(driver_ref) > 0) {
            driver_lock.ref_num_for_sysserv_reboot = atoi(driver_ref);
            driver_lock.ref = atoi(driver_ref);
            LOGD("!!!!!!!!!! ref is %d", driver_lock.ref);
        }
    }

    if(driver_lock.ref <= 0){
        LOGD("driver not loaded");
        return 0;
    }

    /*
     * If the property says the driver is loaded, check to
     * make sure that the property setting isn't just left
     * over from a previous manual shutdown or a runtime
     * crash.
     */
    if ((proc = fopen(MODULE_FILE, "r")) == NULL) {
        LOGW("Could not open %s: %s", MODULE_FILE, strerror(errno));
        return 0;
    }
    while ((fgets(line, sizeof(line), proc)) != NULL) {
        if (strncmp(line, DRIVER_MODULE_TAG, strlen(DRIVER_MODULE_TAG)) == 0) {
            fclose(proc);
            return 1;
        }
    }
    fclose(proc);
    property_set(DRIVER_PROP_NAME, "unloaded");
    LEAVE();

    return 0;
}

static int check_wlan_chip(const char *tag, int secs)
{
#define DRIVER_LIVE_FILE    "/proc/wlan_chip"   
    char *line;
    int line_size;
    FILE *proc;
    int delay;

    line_size = strlen(tag) + 1;
    line = malloc(line_size);
    secs *= 10;
    //setvbuf(proc, NULL, _IONBF, 0);
    for (delay = 0; delay < secs; delay++) {

        if ((proc = fopen(DRIVER_LIVE_FILE, "r")) == NULL) {
            LOGW("Error : %s: %s", DRIVER_LIVE_FILE, strerror(errno));
            free(line);
            return 0;
        }
        
        while ((fgets(line, line_size, proc)) != NULL) {
            if (strncmp(line, tag, line_size) == 0) {
                fclose(proc);
                free(line);
                return 1;
            }
            printf(line);
        }

        fclose(proc);
        //fseek(proc, 0, SEEK_SET);
        //fflush(proc);
        usleep(100*1000);
    }
    
    free(line);
    return 0;
}

static int wlan_load_driver(void)
{
    char driver_status[PROPERTY_VALUE_MAX];
    char property_value[PROPERTY_VALUE_MAX];

    ENTER();
    pthread_mutex_lock(&driver_lock.lock);

    if (check_driver_loaded()) {
        if (driver_lock.ref_num_for_sysserv_reboot > 0) {
            driver_lock.ref_num_for_sysserv_reboot --;
        } else {
            driver_lock.ref++;
        }
        sprintf(property_value, "%d", driver_lock.ref);
        property_set(DRIVER_REF_PROP_NAME, property_value);
        LOGD("wlan driver has been loaded, driver_ref = %d\n",driver_lock.ref);
        pthread_mutex_unlock(&driver_lock.lock);
        return 0;
    }
    if (insmod(DRIVER_MODULE_PATH) < 0) {
        LOGD("in %s, insmod fails", __FUNCTION__);
        pthread_mutex_unlock(&driver_lock.lock);
        return -1;
    } else {
        // Set baud rate and make BT enter deep sleep 
        property_set("ctl.start", "hciattach_dtr");
    }

    //modified by derek.
    //we need not firmware_loader, because init can process sysfs firmware event in userspace.
    //and in kernel space, wifi driver will download firmware to card.
    //therefore, without wifi driver notification, userspace can not make sure when
    //firmware download is over.
    //current solution: let program sleep 1.5 sec.
#if 0
    property_set("ctl.start", FIRMWARE_LOADER);
    while (count-- > 0) {
        usleep(500000);
        if (property_get(DRIVER_PROP_NAME, driver_status, NULL)) {
            if (strcmp(driver_status, "ok") == 0)
                return 0;
            else if (strcmp(DRIVER_PROP_NAME, "failed") == 0)
                return -1;
        }
    }
    property_set(DRIVER_PROP_NAME, "timeout");
#else
    if (check_wlan_chip("driver=live", 10)) {
        property_set(DRIVER_PROP_NAME, "ok");
        driver_lock.ref = 1;
        sprintf(property_value, "%d", driver_lock.ref);
        property_set(DRIVER_REF_PROP_NAME, property_value);
    } else {
        property_set(DRIVER_PROP_NAME, "timeout");
    }
#endif
    pthread_mutex_unlock(&driver_lock.lock);
    LEAVE();
    return 0;
}

static int wlan_unload_driver(void)
{
    char property_value[PROPERTY_VALUE_MAX];

    ENTER();
    pthread_mutex_lock(&driver_lock.lock);
    if (!check_driver_loaded()) {
        driver_lock.ref = 0;
        sprintf(property_value, "%d", driver_lock.ref);
        property_set(DRIVER_REF_PROP_NAME, property_value);
        LOGD("in %s, driver not loaded", __FUNCTION__);
        pthread_mutex_unlock(&driver_lock.lock);
        return 0;
    }
    driver_lock.ref--;
    if(driver_lock.ref < 0)
        driver_lock.ref = 0;
    sprintf(property_value, "%d", driver_lock.ref);
    property_set(DRIVER_REF_PROP_NAME, property_value);

    if(driver_lock.ref > 0){
        pthread_mutex_unlock(&driver_lock.lock);
        return 0;
    }

    if (rmmod(DRIVER_MODULE_NAME) == 0) {
        usleep(1000000);
        property_set(DRIVER_PROP_NAME, "unloaded");
        pthread_mutex_unlock(&driver_lock.lock);
        return 0;
    } else {
        driver_lock.ref++;
        sprintf(property_value, "%d", driver_lock.ref);
        property_set(DRIVER_REF_PROP_NAME, property_value);

        pthread_mutex_unlock(&driver_lock.lock);
        return -1;
    }
}

static const WLAN_ADP_LIB_Functions callbacks = {
    wlan_load_driver,
    wlan_unload_driver,
    wpa_driver_vendor_driver_cmd,
    wpa_driver_vendor_event_custom,
    wpa_driver_init_vendor_data,
    wpa_driver_deinit_vendor_data
};

const WLAN_ADP_LIB_Functions *WLAN_ADP_LIB_Init(void)
{
    ENTER();
    return &callbacks;
}
