/*
 * Copyright 2008, Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>

#include "hardware_legacy/wifi.h"
#include "libwpa_client/wpa_ctrl.h"

#define LOG_TAG "WifiHW"
#include "cutils/log.h"
#include "cutils/memory.h"
#include "cutils/misc.h"
#include "cutils/properties.h"
#include "private/android_filesystem_config.h"

//added by derek
#define WIFI_DEBUG

#ifdef WIFI_DEBUG 
#define ENTER() LOGD("enter into %s", __FUNCTION__);
#define LEAVE() LOGD("leave from %s", __FUNCTION__);
#else
#define ENTER() do { } while (0)
#define LEAVE() do { } while (0)
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

static struct wpa_ctrl *wapi_ctrl_conn;
static struct wpa_ctrl *wapi_monitor_conn;
//this global static varible describes if we need reset socket or not.
static int need_reset_sock = FALSE;

static const char ctrl_intf_path[]           = "/data/system/wpa_supplicant/eth0";

extern int do_dhcp();
extern int ifc_init();
extern void ifc_close();
extern char *dhcp_lasterror();
extern void get_dhcp_info();
extern int init_module(void *, unsigned long, const char *);
extern int delete_module(const char *, unsigned int);

static char iface[PROPERTY_VALUE_MAX];
// TODO: use new ANDROID_SOCKET mechanism, once support for multiple
// sockets is in

static const char IFACE_DIR[]           = "/data/system/wapi_supplicant";
static const char DRIVER_MODULE_NAME[]  = "sd8688";
static const char DRIVER_MODULE_TAG[]   = "sd8688 ";
static const char DRIVER_MODULE_PATH[]  = "/opl/modules/sd8688.ko";
static const char FIRMWARE_LOADER[]     = "wlan_loader";
static const char DRIVER_PROP_NAME[]    = "wapi.driver.status";
static const char DRIVER_REF_PROP_NAME[]= "wapi.driver.ref"; //added by li yanbing
static const char SUPPLICANT_NAME[]     = "wapi_supplicant";
static const char SUPP_PROP_NAME[]      = "init.svc.wapi_supplicant";
static const char SUPP_CONFIG_TEMPLATE[]= "/system/etc/wapi/wapi_supplicant.conf";
static const char SUPP_CONFIG_FILE[]    = "/data/misc/wapi/wapi_supplicant.conf";
static const char SUPP_LOG_FILE[]    = "/data/misc/wapi/wapi_log.txt";
static const char SUPP_LOG_PRE_FILE[]    = "/data/misc/wapi/wapi_log_pre.txt";
static const char SUPP_CLI_IFACE_DIR[]    = "/data/misc/wapi/sockets";
static const char MODULE_FILE[]         = "/proc/modules";


static int insmod(const char *filename)
{
    void *module;
    unsigned int size;
    int ret;

    ENTER();
    module = load_file(filename, &size);
    if (!module)
        return -1;

    ret = init_module(module, size, "");

    free(module);
    LEAVE();

    return ret;
}

static int rmmod(const char *modname)
{
    int ret = -1;
    int maxtry = 10;

    ENTER();
    while (maxtry-- > 0) {
        ret = delete_module(modname, O_NONBLOCK | O_EXCL);
        if (ret < 0 && errno == EAGAIN)
            usleep(500000);
        else
            break;
    }

    if (ret != 0)
        LOGD("Unable to unload driver module \"%s\": %s\n",
             modname, strerror(errno));
    LEAVE();

    return ret;
}

int wapi_do_dhcp_request(int *ipaddr, int *gateway, int *mask,
                    int *dns1, int *dns2, int *server, int *lease) {
    ENTER();
    /* For test driver, always report success */
    if (strcmp(iface, "sta") == 0)
        return 0;

    if (ifc_init() < 0)
        return -1;

    if (do_dhcp(iface) < 0) {
        ifc_close();
        return -1;
    }
    ifc_close();
    get_dhcp_info(ipaddr, gateway, mask, dns1, dns2, server, lease);
    LEAVE();

    return 0;
}

const char *wapi_get_dhcp_error_string() {
    return dhcp_lasterror();
}

static int check_driver_loaded() {
    char driver_status[PROPERTY_VALUE_MAX];
    FILE *proc;
    char line[sizeof(DRIVER_MODULE_TAG)+10];

    ENTER();
    if (!property_get(DRIVER_PROP_NAME, driver_status, NULL)
            || strcmp(driver_status, "ok") != 0) {
        return 0;  /* driver not loaded */
    }
    /*
     * If the property says the driver is loaded, check to
     * make sure that the property setting isn't just left
     * over from a previous manual shutdown or a runtime
     * crash.
     */
    if ((proc = fopen(MODULE_FILE, "r")) == NULL) {
        LOGW("Could not open %s: %s", MODULE_FILE, strerror(errno));
        return 0;
    }
    while ((fgets(line, sizeof(line), proc)) != NULL) {
        if (strncmp(line, DRIVER_MODULE_TAG, strlen(DRIVER_MODULE_TAG)) == 0) {
            fclose(proc);
            return 1;
        }
    }
    fclose(proc);
    property_set(DRIVER_PROP_NAME, "unloaded");
    LEAVE();

    return 0;
}
//added by li yanbing
static void increase_driver_ref()
{
    char driver_ref[PROPERTY_VALUE_MAX];
    int ref;

    if (property_get(DRIVER_REF_PROP_NAME, driver_ref, NULL)){
        ref = atoi(driver_ref);
        ref++;
        memset(driver_ref, 0, sizeof(driver_ref));
        sprintf(driver_ref, "%d", ref);
        property_set(DRIVER_REF_PROP_NAME, driver_ref);
    }
}
//added by li yanbing
static int decrease_driver_ref()
{
    char driver_ref[PROPERTY_VALUE_MAX];
    int ref = 0;

    if (property_get(DRIVER_REF_PROP_NAME, driver_ref, NULL)){
        ref = atoi(driver_ref);
        ref--;
        if(ref < 0){
            ref = 0;
        }
        memset(driver_ref, 0, sizeof(driver_ref));
        sprintf(driver_ref, "%d", ref);
        property_set(DRIVER_REF_PROP_NAME, driver_ref);
    }
    return ref;
}

int wapi_load_driver()
{
    char driver_status[PROPERTY_VALUE_MAX];
    int count = 40; /* wait at most 20 seconds for completion */

    ENTER();
    if (check_driver_loaded()) {
        increase_driver_ref();
        return 0;
    }
    if (insmod(DRIVER_MODULE_PATH) < 0) {
        LOGD("in wifi_loader_driver function, insmod fails\n");
        return -1;
    }

    //modified by derek.
    //we need not firmware_loader, because init can process sysfs firmware event in userspace.
    //and in kernel space, wifi driver will download firmware to card.
    //therefore, without wifi driver notification, userspace can not make sure when
    //firmware download is over.
    //current solution: let program sleep 1.5 sec.
#if 0
    property_set("ctl.start", FIRMWARE_LOADER);
    while (count-- > 0) {
        usleep(500000);
        if (property_get(DRIVER_PROP_NAME, driver_status, NULL)) {
            if (strcmp(driver_status, "ok") == 0)
                return 0;
            else if (strcmp(DRIVER_PROP_NAME, "failed") == 0)
                return -1;
        }
    }
    property_set(DRIVER_PROP_NAME, "timeout");
#else
    usleep(1500000);
    property_set(DRIVER_PROP_NAME, "ok");
    property_set(DRIVER_REF_PROP_NAME, "1");
#endif
    LEAVE();

    return 0;
}

int wapi_unload_driver()
{
    ENTER();

    if (!check_driver_loaded()) {
        LOGD("in wifi_unload_driver function, driver not loaded\n");
        return 0;
    }
    if(decrease_driver_ref()){
        return 0;
    }

    if (rmmod(DRIVER_MODULE_NAME) == 0) {
        usleep(1000000);
        property_set(DRIVER_PROP_NAME, "unloaded");
        return 0;
    } else
        return -1;
}

static int control_supplicant(int startIt)
{
    char supp_status[PROPERTY_VALUE_MAX] = {'\0'};
    const char *ctrl_prop = (startIt ? "ctl.start" : "ctl.stop");
    const char *desired_status = (startIt ? "running" : "stopped");
    int count = 20; /* wait at most 20 seconds for completion */

    ENTER();
    if (property_get(SUPP_PROP_NAME, supp_status, NULL)
        && strcmp(supp_status, desired_status) == 0) {
        LOGD("in control_supplicant, supplicant is already running\n");
        return 0;  /* supplicant already running */
    }
    LOGD("in control_supplicant, supplicant will be launched\n");
    property_set(ctrl_prop, SUPPLICANT_NAME);

    while (count-- > 0) {
        usleep(1000000);
        if (property_get(SUPP_PROP_NAME, supp_status, NULL)) {
            if (strcmp(supp_status, desired_status) == 0)
                return 0;
        }
    }
    LEAVE();

    return -1;

#if 0
    char supp_status[PROPERTY_VALUE_MAX] = {'\0'};

    property_get(SUPP_PROP_NAME, supp_status, NULL); 
    if(startIt){
        if (strcmp(supp_status, "running") != 0){
            system("wapic -imlan0 -c /data/misc/wifi/wapi.conf &");
            property_set(SUPP_PROP_NAME, "running");
        }
    }else{
        if (strcmp(supp_status, "running") == 0){
            system("killall wapic");
            property_set(SUPP_PROP_NAME, "stopped");
        }
    }
    return 0;
#endif
}

static int ensure_config_file_exists()
{
    char buf[2048];
    int srcfd, destfd;
    int nread;

    ENTER();
    if (access(SUPP_CONFIG_FILE, R_OK|W_OK) == 0) {
        return 0;
    } else if (errno != ENOENT) {
        LOGE("Cannot access \"%s\": %s", SUPP_CONFIG_FILE, strerror(errno));
        return -1;
    }

    srcfd = open(SUPP_CONFIG_TEMPLATE, O_RDONLY);
    if (srcfd < 0) {
        LOGE("Cannot open \"%s\": %s", SUPP_CONFIG_TEMPLATE, strerror(errno));
        return -1;
    }

    destfd = open(SUPP_CONFIG_FILE, O_CREAT|O_WRONLY, 0660);
    if (destfd < 0) {
        close(srcfd);
        LOGE("Cannot create \"%s\": %s", SUPP_CONFIG_FILE, strerror(errno));
        return -1;
    }

    while ((nread = read(srcfd, buf, sizeof(buf))) != 0) {
        if (nread < 0) {
            LOGE("Error reading \"%s\": %s", SUPP_CONFIG_TEMPLATE, strerror(errno));
            close(srcfd);
            close(destfd);
            unlink(SUPP_CONFIG_FILE);
            return -1;
        }
        write(destfd, buf, nread);
    }

    close(destfd);
    close(srcfd);

    if (chown(SUPP_CONFIG_FILE, AID_SYSTEM, AID_WIFI) < 0) {
        LOGE("Error changing group ownership of %s to %d: %s",
             SUPP_CONFIG_FILE, AID_WIFI, strerror(errno));
        unlink(SUPP_CONFIG_FILE);
        return -1;
    }
    LEAVE();

    return 0;
}

static void prepare_for_supplicant()
{
    DIR * dp = NULL;
    struct dirent * dirp = NULL;
    char iface_name[128];

    ENTER();
    //delete wpa_log_pre.txt file
    if (access(SUPP_LOG_PRE_FILE, F_OK|W_OK) == 0) {
        if (unlink(SUPP_LOG_PRE_FILE) == 0)
            LOGD("delete %s successfully", SUPP_LOG_PRE_FILE);
        else
            LOGD("delete %s unsuccessfully", SUPP_LOG_PRE_FILE);
    }

    //rename wpa_log.txt file
    if (access(SUPP_LOG_FILE, F_OK|W_OK) == 0) {
        if (rename(SUPP_LOG_FILE, SUPP_LOG_PRE_FILE) == 0)
            LOGD("rename %s successfully", SUPP_LOG_FILE);
        else
            LOGD("rename %s unsuccessfully", SUPP_LOG_FILE);
    }

    //delete all socket local file in /data/misc/wifi/sockets
    if ((dp = opendir(SUPP_CLI_IFACE_DIR)) == NULL) {
        LOGD("open dir %s unsuccessfully\n", SUPP_CLI_IFACE_DIR);
        return;
    }

    while ((dirp = readdir(dp)) != NULL)
    {
        if(strncmp(dirp->d_name, "wpa_ctrl_", 9) == NULL)
        {
            sprintf(iface_name,"%s/%s", SUPP_CLI_IFACE_DIR, dirp->d_name);
            if (unlink(iface_name) == 0)
                LOGD("delete %s successfully\n", dirp->d_name);
            else
                LOGD("delete %s unsuccessfully\n", dirp->d_name);
        }
    }
    closedir(dp);
 
    LEAVE();
    return;
}

int wapi_start_supplicant()
{
    ENTER();
    /* Before starting the daemon, make sure its config file exists */
    if (ensure_config_file_exists() < 0) {
        LOGE("Wi-Fi will not be enabled");
        return -1;
    }

    //prepare_for_supplicant();
    return control_supplicant(1);
}

int wapi_stop_supplicant()
{
    ENTER();

    return control_supplicant(0);
}

int wapi_connect_to_supplicant()
{
    char ifname[256];
    static int cleaned_up = 0;
    ENTER();
#if 0
    property_get("wifi.interface"IFACE_DIR, iface, "sta");

    if (access(IFACE_DIR, F_OK) == 0) {
        snprintf(ifname, sizeof(ifname), "%s/%s", IFACE_DIR, iface);
    } else {
        strlcpy(ifname, iface, sizeof(ifname));
    }
    LOGD("in wifi_connect_to_supplicant, ifname is %s", ifname);
#endif
    wapi_ctrl_conn = wpa_ctrl_open(ctrl_intf_path);
    if (wapi_ctrl_conn == NULL) {
        LOGD("Unable to open connection to supplicant on \"%s\": %s",
             ifname, strerror(errno));
        /*
         * errno == ENOENT means the supplicant daemon isn't
         * running. Take this opportunity to clear out any
         * stale socket files that might be left over. Note
         * there's a possible race with the command line client
         * trying to connect to the daemon, but it would require
         * that the supplicant be started and the command line
         * client connect to it during the window between the
         * error check and the removal of the files. And in
         * any event, the remedy is that the user would simply
         * have to run the command line program again.
         */
        if (!cleaned_up && (errno == ENOENT || errno == EADDRINUSE)) {
            cleaned_up = 1; /* do this just once */
            wpa_ctrl_cleanup();
        }
        return -1;
    }
#if 0
    monitor_conn = wpa_ctrl_open(ifname);
    if (monitor_conn == NULL) {
        wpa_ctrl_close(ctrl_conn);
        ctrl_conn = NULL;
        return -1;
    }
    if (wpa_ctrl_attach(monitor_conn) != 0) {
        wpa_ctrl_close(monitor_conn);
        wpa_ctrl_close(ctrl_conn);
        ctrl_conn = monitor_conn = NULL;
        return -1;
    }
#endif
    LEAVE();

    return 0;
}

static int wifi_send_command(struct wpa_ctrl *ctrl, const char *cmd, char *reply, size_t *reply_len)
{
    int ret;
    static int try_num = 0;
#define WIFI_SEND_FAIL_TRY_MAX_NUM 5

    LOGD("enter into wifi_send_command, cmd is %s", cmd);

    if (wapi_ctrl_conn == NULL || need_reset_sock == TRUE) {
        LOGV("Not connected to wpa_supplicant - \"%s\" command dropped.\n", cmd);
        return -1;
    }
    ret = wpa_ctrl_request(ctrl, cmd, strlen(cmd), reply, reply_len, NULL);
    LOGD("in wifi_send_command, ret is %d, reply is %s", reply_len, reply);
    if (ret == -3) {
        if (++ try_num >= WIFI_SEND_FAIL_TRY_MAX_NUM) {
        	//set need_reset_sock to TRUE to notify event thread to process for it.
        	LOGD("send fails %d times, set need_reset_sock to TRUE", WIFI_SEND_FAIL_TRY_MAX_NUM);
    		need_reset_sock = TRUE;
    		try_num = 0;
        }
        return -1;
    } else if (ret == -2) {
        LOGD("'%s' command timed out.\n", cmd);
        return -2;
    } else if (ret < 0 || strncmp(reply, "FAIL", 4) == 0) {
        return -1;
    }
 
    //reset try_num to zero if it sends successfully.
    try_num = 0;
    if (strncmp(cmd, "PING", 4) == 0) {
        reply[*reply_len] = '\0';
    }

    return 0;
}

int wapi_wait_for_event(char *buf, size_t buflen)
{
    size_t nread = buflen - 1;
    int fd;
    fd_set rfds;
    int result;
    struct timeval tval;
    struct timeval *tptr;
    
    //ENTER();
    if (wapi_monitor_conn == NULL) {
        LOGD("monitor_conn is NULL");
        sleep(1);
        return 0;
    }

    while (1) {
    	if (need_reset_sock == TRUE) {
        	LOGD("try to reset local socket by default for wifi feature robustness");
	        wifi_close_supplicant_connection();
        	wifi_connect_to_supplicant();
        	need_reset_sock = FALSE;
    	}
        
    	if(wpa_ctrl_pending_timeout(wapi_monitor_conn, 2, 0) > 0) {
    		//need call recv func
        	result = wpa_ctrl_recv(wapi_monitor_conn, buf, &nread);
        	break;
    	}
    }

    if (result < 0) {
        LOGD("wpa_ctrl_recv failed: %s\n", strerror(errno));
        return -1;
    }
    buf[nread] = '\0';
    /* LOGD("wait_for_event: result=%d nread=%d string=\"%s\"\n", result, nread, buf); */
    /* Check for EOF on the socket */
    if (result == 0 && nread == 0) {
        /* Fabricate an event to pass up */
        LOGD("Received EOF on supplicant socket\n");
        strncpy(buf, WPA_EVENT_TERMINATING " - signal 0 received", buflen-1);
        buf[buflen-1] = '\0';
        return strlen(buf);
    }
    /*
     * Events strings are in the format
     *
     *     <N>CTRL-EVENT-XXX 
     *
     * where N is the message level in numerical form (0=VERBOSE, 1=DEBUG,
     * etc.) and XXX is the event name. The level information is not useful
     * to us, so strip it off.
     */
    if (buf[0] == '<') {
        char *match = strchr(buf, '>');
        if (match != NULL) {
            nread -= (match+1-buf);
            memmove(buf, match+1, nread+1);
        }
    }
    //LEAVE();

    return nread;
}

void wapi_close_supplicant_connection()
{
    ENTER();
    if (wapi_ctrl_conn != NULL) {
        wpa_ctrl_close(wapi_ctrl_conn);
        wapi_ctrl_conn = NULL;
    }
#if 0
    if (monitor_conn != NULL) {
        wpa_ctrl_close(monitor_conn);
        monitor_conn = NULL;
    }
#endif
    LEAVE();
}

int wapi_command(const char *command, char *reply, size_t *reply_len)
{
    //ENTER();
    return wifi_send_command(wapi_ctrl_conn, command, reply, reply_len);
}


int wapi_set_net_config(const char *ssid, const char *key_mgmt, const char *psk, const char *cert)
{
    FILE *fd;
    int res = 0;

    if(ssid == NULL || key_mgmt == NULL)
        return -1;
    if ((fd = fopen(SUPP_CONFIG_FILE, "w")) == NULL)    
        return -1;
    fprintf(fd, "SSID=%s\n",ssid);
    if(strncmp(key_mgmt, "open", 4) == 0){
        fprintf(fd, "WAPI=3\n");
    }else if(strncmp(key_mgmt, "psk", 3) == 0){
        fprintf(fd, "WAPI=7\n");
        if(psk == NULL){
            res = -1;
            goto exit;
	}
	if( psk[0] == '"' ){
		fprintf(fd, "PSK_KEY_TYPE=0\n");
		fprintf(fd, "PSK=%s\n",++psk);
	}
	else {
		fprintf(fd, "PSK_KEY_TYPE=1\n");
		fprintf(fd, "PSK=%s\n",psk);
	}
    }else if(strncmp(key_mgmt, "cert", 4) == 0){
        fprintf(fd, "WAPI=11\n");
        if(cert == NULL){
            res = -1;
            goto exit;
	}
        fprintf(fd, "CERT_NAME=%s\n",cert);
    }
    fprintf(fd, "CERT_INDEX=1\n");
    fprintf(fd, "CERT_STATUS=1\n");
    fprintf(fd, "WLAN_INTERFACE=eth0\n");
    
exit:   
    fclose(fd);
    return res;
}

/* parameters: none
 * return:
 * "SSID=
 * KEY_MGMT=
 * PSK=
 * CERT="
 */
int wapi_get_net_config(char *buf, size_t buf_len)
{
    FILE *fd;
    char line[256];
    char *pos;
    char *index;
    int  i = 0;

    if(buf == NULL || buf_len == 0)
        return -1;
    if ((fd = fopen(SUPP_CONFIG_FILE, "r")) == NULL)    
        return -1;
    memset(buf, 0, buf_len);
    pos = buf;
    while(!feof(fd)){
        fgets(line, 256, fd);
        if((index = strstr(line, "SSID=")) != NULL && index == line){
	    while( line[i] != '\n' && i <= 256 )
		i++;
	    if( line[i] == '\n' )
		line[i] = '\0';		
            pos += sprintf(pos, "%s\t", line + 5);
        }else if((index = strstr(line, "WAPI=")) != NULL && index == line){
            i = 0;
	    while( line[i] != '\n' && i <= 256 )
		i++;
	    if( line[i] == '\n' )
		line[i] = '\0';		
            int wapi;
            wapi = atoi(line + sizeof("WAPI=") - 1);
	    if(wapi == 3 ){
		pos += sprintf(pos, "open\t");
	    }else if(wapi == 7){
                pos += sprintf(pos, "psk\t");
            }else if(wapi == 11){
                pos += sprintf(pos, "cert\t");
            }
        }else if((index = strstr(line, "PSK=")) != NULL && index == line){
            i = 0;
	    while( line[i] != '\n' && i <= 256 )
		i++;
	    if( line[i] == '\n' )
		line[i] = '\0';		
            pos += sprintf(pos, "%s", line + 4);
        }else if((index = strstr(line, "CERT_NAME=")) != NULL && index == line){
            i = 0;
	    while( line[i] != '\n' && i <= 256 )
		i++;
	    if( line[i] == '\n' )
		line[i] = '\0';		
            pos += sprintf(pos, "%s", line+10);
        }
    }
    fclose(fd);
    return 0;   
}

