/** @file wapic/common.c
 *  @brief This file contains functions for common
 *
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>

#include "include/common.h"

/** Debug level */
int wpa_debug_level = MSG_INFO;
/** Debug flag to show keys */
int wpa_debug_show_keys = 0;
/** Debug flag to show timestamp */
int wpa_debug_timestamp = 0;

/**
 * @brief Converts a hexadecimal character to integer
 *
 * @param c     Hexadecimal character
 *
 * @return      Corresponding integer value
 */
int hex2num(char c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	if (c >= 'A' && c <= 'F')
		return c - 'A' + 10;
	return -1;
}

/**
 * @brief Converts an hexadecimal character to byte value
 *
 * @param hex   Hexadecimal character
 *
 * @return      Corresponding byte value
 */
static int hex2byte(const char *hex)
{
	int a, b;
	a = hex2num(*hex++);
	if (a < 0)
		return -1;
	b = hex2num(*hex++);
	if (b < 0)
		return -1;
	return (a << 4) | b;
}

/**
 * @brief Converts a hexadecimal string to binary
 *
 * @param hex   Hexadecimal string
 * @param buf   Output buffer for binary string
 * @param len   Length of the string
 *
 * @return      0 if success, else -1
 */
int hexstr2bin(const char *hex, u8 *buf, size_t len)
{
	unsigned int i;
	int a;
	const char *ipos = hex;
	u8 *opos = buf;

	for (i = 0; i < len; i++) {
		a = hex2byte(ipos);
		if (a < 0)
			return -1;
		*opos++ = a;
		ipos += 2;
	}
	return 0;
}

/**
 * @brief Converts a relative path to absolute
 *
 * @param rel_path  Relative path string
 *
 * @return          Absolute path string
 */
char * rel2abs_path(const char *rel_path)
{
	char *buf = NULL, *cwd, *ret;
	size_t len = 128, cwd_len, rel_len, ret_len;

	if (rel_path[0] == '/')
		return strdup(rel_path);

	for (;;) {
		buf = malloc(len);
		if (buf == NULL)
			return NULL;
		cwd = getcwd(buf, len);
		if (cwd == NULL) {
			free(buf);
			if (errno != ERANGE) {
				return NULL;
			}
			len *= 2;
		} else {
			break;
		}
	}

	cwd_len = strlen(cwd);
	rel_len = strlen(rel_path);
	ret_len = cwd_len + 1 + rel_len + 1;
	ret = malloc(ret_len);
	if (ret) {
		memcpy(ret, cwd, cwd_len);
		ret[cwd_len] = '/';
		memcpy(ret + cwd_len + 1, rel_path, rel_len);
		ret[ret_len - 1] = '\0';
	}
	free(buf);
	return ret;
}

/**
 * @brief Print debug message timestamps
 *
 * @return      None
 */
static void wpa_debug_print_timestamp(void)
{
	struct timeval tv;
	char buf[16];

	if (!wpa_debug_timestamp)
		return;

	gettimeofday(&tv, NULL);
	if (strftime(buf, sizeof(buf), "%b %d %H:%M:%S",
		     localtime((const time_t *) &tv.tv_sec)) <= 0) {
		snprintf(buf, sizeof(buf), "%u", (int) tv.tv_sec);
	}
	printf("%s.%06u: ", buf, (unsigned int) tv.tv_usec);
}

/**
 * @brief Print message
 *
 * @param level     Message level
 * @param fmt       Printf style formatting
 *
 * @return          None
 */
void wpa_printf(int level, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (level >= wpa_debug_level)
		{
		wpa_debug_print_timestamp();
		vprintf(fmt, ap);
		printf("\n");
	}
	va_end(ap);
}

/**
 * @brief Internal hexdump function
 *
 * @param level     Message level
 * @param title     Hexdump title
 * @param buf       Buffer to dump
 * @param len       Length of buffer to dump
 * @param show      Flag to enable/disable message showing
 *
 * @return          None
 */
static void _wpa_hexdump(int level, const char *title, const u8 *buf,
			 size_t len, int show)
{
	size_t i;
	if (level < wpa_debug_level)
		return;
	wpa_debug_print_timestamp();
	printf("%s - hexdump(len=%lu):\n", title, (unsigned long) len);
	if (show) {
		for (i = 0; i < len; i++){
			printf("%02x ", buf[i]);
			if((i+1) %16 == 0){
				printf("\n");
			}
		}
	} else {
		printf(" [REMOVED]");
	}
	printf("\n");
}

/**
 * @brief Hexdump function
 *
 * @param level     Message level
 * @param title     Hexdump title
 * @param buf       Buffer to dump
 * @param len       Length of buffer to dump
 *
 * @return          None
 */
void wpa_hexdump(int level, const char *title, const u8 *buf, size_t len)
{
	_wpa_hexdump(level, title, buf, len, 1);
}

/**
 * @brief Hexdump key
 *
 * @param level     Message level
 * @param title     Hexdump title
 * @param buf       Buffer to dump
 * @param len       Length of buffer to dump
 *
 * @return          None
 */
void wpa_hexdump_key(int level, const char *title, const u8 *buf, size_t len)
{
	_wpa_hexdump(level, title, buf, len, wpa_debug_show_keys);
}

/**
 * @brief Internal hexdump ASCII
 *
 * @param level     Message level
 * @param title     Hexdump title
 * @param buf       Buffer to dump
 * @param len       Length of buffer to dump
 * @param show      Flag to enable/disable message showing
 *
 * @return          None
 */
static void _wpa_hexdump_ascii(int level, const char *title, const u8 *buf,
			       size_t len, int show)
{
	unsigned int i;
	unsigned int llen;
	const u8 *pos = buf;
	const unsigned  int line_len = 16;

	if (level < wpa_debug_level)
		return;
	wpa_debug_print_timestamp();
	if (!show) {
		printf("%s - hexdump_ascii(len=%lu): [REMOVED]\n",
		       title, (unsigned long) len);
		return;
	}
	printf("%s - hexdump_ascii(len=%lu):\n", title, (unsigned long) len);
	while (len) {
		llen = len > line_len ? line_len : len;
		printf("    ");
		for (i = 0; i < llen; i++)
			printf(" %02x", pos[i]);
		for (i = llen; i < line_len; i++)
			printf("   ");
		printf("   ");
		for (i = 0; i < llen; i++) {
			if (isprint(pos[i]))
				printf("%c", pos[i]);
			else
				printf("_");
		}
		for (i = llen; i < line_len; i++)
			printf(" ");
		printf("\n");
		pos += llen;
		len -= llen;
	}
}

/**
 * @brief Hexdump ASCII
 *
 * @param level     Message level
 * @param title     Hexdump title
 * @param buf       Buffer to dump
 * @param len       Length of buffer to dump
 *
 * @return          None
 */
void wpa_hexdump_ascii(int level, const char *title, const u8 *buf, size_t len)
{
	_wpa_hexdump_ascii(level, title, buf, len, 1);
}

/**
 * @brief Hexdump ASCII key
 *
 * @param level     Message level
 * @param title     Hexdump title
 * @param buf       Buffer to dump
 * @param len       Length of buffer to dump
 *
 * @return          None
 */
void wpa_hexdump_ascii_key(int level, const char *title, const u8 *buf,
			   size_t len)
{
	_wpa_hexdump_ascii(level, title, buf, len, wpa_debug_show_keys);
}


#ifdef CONFIG_NATIVE_WINDOWS
/** Epoch file time */
#define EPOCHFILETIME (116444736000000000ULL)

/**
 * brief Get system time
 *
 * @param tv    Time value
 * @param tz    Time zone
 *
 * @return      0
 */
int gettimeofday(struct timeval *tv, struct timezone *tz)
{
	FILETIME ft;
	LARGE_INTEGER li;
	ULONGLONG t;

	GetSystemTimeAsFileTime(&ft);
	li.LowPart = ft.dwLowDateTime;
	li.HighPart = ft.dwHighDateTime;
	t = (li.QuadPart - EPOCHFILETIME) / 10;
	tv->tv_sec = (long) (t / 1000000);
	tv->tv_usec = (long) (t % 1000000);

	return 0;
}
#endif /* CONFIG_NATIVE_WINDOWS */
