/** @file config.c
  * @brief This file contains functions for config
 *
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <stdlib.h>
#include "include/config.h"
#include "include/wapi_asue.h"

/** Separation mark */
#define SEP_Q_MARK	'"'

/**
 * @brief Parse a line to get property
 *
 * @param type  Separator type
 * @param line  Property line
 * @param prop  Property data structure to fill
 *
 * @return      1 if new property found, 0 otherwise
 */
static int get_pr (char type, char *line, prop_data *prop)
{
    char *index = NULL;	
    char *keyptr = NULL;

    if (line[0] != '#')
    {
        // modified by kWlW, separator changed to this  vvv  from "="
        keyptr = (char *)strstr(line, "setenv ");	
		//finds the first occurrence of the substring 'setenv' in line
        if (keyptr != NULL) line = keyptr + 7;   
		/* Add size of 'setenv ' string */
		index =  strstr (line, type == SEP_EQUAL ? "=" : type == SEP_SPACE ? " " : "\t");	
        if (((char *)index) != NULL)
        {
            *((char *) index) = '\0';
            if ((prop->key = (char *)malloc (strlen (line) + 1)) != NULL &&
                (prop->value =(char *) malloc (strlen ((char *) (index + 1)) + 1)) != NULL)
            {
                strcpy (prop->key, line);
                strcpy (prop->value, (char *) (index + 1));
                return 1;
            }
        }
    }
    return 0;
}

/**
 * @brief Load property
 *
 * @param type          Separator type
 * @param file_name     Propoerty file
 * @param properties    Property data to fill
 *
 * @return              Number of properties loaded
 */
int load_prop (char type, char *file_name, prop_data properties[])
{
	FILE *fid;
	char line[1024];
	int prop_count = 0;

	if ((fid = fopen (file_name, "r")) == NULL)    return 0;
	while (!feof (fid))
	{
		fgets(line, 255, fid);
		if(feof(fid)) break;
		if (strchr(line, '\n'))  *(strchr(line, '\n')) = '\0';
		if (get_pr (type, line, (prop_data *) & properties[prop_count]))
		{
			prop_count++;
		}
	}
	fclose (fid);
	return prop_count;
}

/**
 * @brief Strip a string
 *
 * @param string    String to strip
 * @param delimiter Character to strip
 *
 * @return          Stripped string
 */
char *strip (char *string, char delimiter)
{
  register int x, y;

  y = strlen (string);
  x = 0;
  while ((x < y) && (string[x] == delimiter))
    {
	  x++;
    };
  while ((y > 0) && (string[y - 1] == delimiter))
    {
      y--;
    };
  string[y] = 0;
  return string + x;
}

/**
 * @brief Get property
 *
 * @param key           Key of property to get
 * @param result        Retreived property
 * @param properties    Property database
 * @param count         Property count
 *
 * @return              Retrieved property if successful, NULL otherwise
 */
char *get_prop (char *key, char *result, prop_data properties[], int count)
{
  int index;
  char tmp_str[256] = "";
  for (index = 0; index < count; index++)
    {
      if (strcmp (properties[index].key, key) == 0)
	{
	  strcpy(tmp_str, properties[index].value);	
	  strcpy (result, strip(tmp_str, SEP_Q_MARK));
	  return result;
	}
    }
  return NULL;
}

/**
 * @brief Converts hexadecimal character to integer
 *
 * @param c     Hexadecimal character to convert
 *
 * @result      Corresponding integer value
 */
int hex2int(char c)
{
	if (c >= '0' && c <= '9')
		return (c - '0');
	if (c >= 'a' && c <= 'f')
		return (c - 'a' + 10);
	if (c >= 'A' && c <= 'F')
		return (c - 'A' + 10);
	return -1;
}

/**
 * @brief Converts string to byte
 *
 * @param str       String to convert
 * @param len       String length
 * @param byte_out  Output buffer for byte values
 *
 * @return          0 if successful, else -1
 */
int str2byte( unsigned char *str, int len,  char *byte_out)
{
	int i, val, val2;
	unsigned char *pos = str;
	
	for (i = 0; i < len/2; i++) {
		val = hex2int(*pos++);
		if (val < 0)
			return -1;
		val2 = hex2int(*pos++);
		if (val2 < 0)
			return -1;
		byte_out[i] = (val * 16 + val2) & 0xff;
	}

	return 0;
}

/**
 * @brief Allocate a buffer
 *
 * @param len   Length of buffer to allocate
 *
 * @return      Pointer to allocated buffer, NULL if failure
 */
void *get_buffer(int len)
{
	char *buffer=NULL;
	buffer = (char *)malloc(len);
	if(buffer)
		memset(buffer, 0, len);
	else
		buffer = NULL;
	return buffer;
}

/**
 * @brief Free a previously allocated buffer
 *
 * @param buffer    Buffer to free
 * @param len       Length of the buffer
 *
 * @return          NULL
 */
void *free_buffer(void *buffer, int len)
{
	char *tmpbuf = (char *)buffer;

	if(tmpbuf != NULL)
	{
		memset(tmpbuf, 0, len);
		free(tmpbuf);
		return NULL;
	}
	else
		return NULL;
}

/**
 * @brief Free properties
 *
 * @param properties   Property database
 * @param count         Number of properties to free
 *
 * @return              1
 */
int free_prop (prop_data properties[], int count)
{
  int index;
  for (index = 0; index < count; index++)
    {
      free (properties[index].key);
      free (properties[index].value);
    }
  return 1;
}

/**
 * @brief Save global configuration
 *
 * @param type      Separator type
 * @param filename  Configuration file name
 * @param command   Command to save
 * @param key       Key to save
 * @param value     Value to save
 *
 * @return          1
 */
int save_global_conf(char type, char* filename, char *command, char *key, char *value)
{
	FILE* fidIn, *fidOut;
	char line[1024], tmp_line[1024];
        char tmp_fname[255], *index;
        char separator[2];
        int i=0;
	
	if ((fidIn=fopen(filename,"r"))==NULL) return 0;
	strcpy(tmp_fname,filename);	
	strcat(tmp_fname,".new");
	
        if ((fidOut=fopen(tmp_fname,"w"))==NULL) return 0;

        strcpy(separator,(type==SEP_EQUAL?"=":type==SEP_SPACE?" ":"\t"));
        type=separator[0];

	while(!feof(fidIn))
	{
		fgets(line,1024,fidIn);
		strcpy(tmp_line,command);
		strcat(tmp_line,key);
                if (((index=strstr(line,tmp_line))!=NULL) && (index==line)) {
					printf("finding\n");
                    i=1;
                    if (!feof(fidIn)) {
                        if (type != '0')
                            fprintf(fidOut,"%s%s%c%s\n",command,key,type,value);
                        else
                            fprintf(fidOut,"%s%s%s\n",command,key,value);
                    }
                } else
                    if (!feof(fidIn)) fprintf(fidOut,"%s",line);
        }

        if (i==0) {
            if (type != '0') { fprintf(fidOut,"%s%s%c%s\n",command,key,type,value); }
            else fprintf(fidOut,"%s%s%s\n",command,key,value);
        }
             
	fclose(fidIn);
	fclose(fidOut);
	unlink(filename);
	rename(tmp_fname,filename);
	return 1;

}

/**
 * @brief Read configuration file
 *
 * @param config_file   Configuration file name
 * @param buf           Buffer to read data into
 * @param buflen        Buffer length
 *
 * @return              Buffer position, -1 if failure
 */
int  wpa_network_read(const char *config_file,char *buf, size_t buflen)
{
	prop_data properties[KEYS_MAX];
	int prop_count=0;
	char *pos, *end;
	char value[256] = {0,};
	
	pos = buf;
	end = buf + buflen;
	wpa_printf(MSG_DEBUG, "Reading configuration file '%s'",
		   config_file);
	prop_count=load_prop(SEP_EQUAL, (char *)config_file,properties);	

	if(prop_count == 0){
		wpa_printf(MSG_ERROR, "no keys and values pair in configuration file '%s'",
			   config_file);
		return -1;
	}
	
	bzero(value, 256);
	get_prop("SSID", value, properties, prop_count);
	pos += snprintf(pos, end - pos, "SSID=%s\n",value);
	
	bzero(value, 256);
	get_prop("WAI", value, properties, prop_count);
	pos += snprintf(pos, end - pos, "WAI=%s\n",value);
	bzero(value, 256);
	get_prop("WPI", value, properties, prop_count);
	pos += snprintf(pos, end - pos, "WPI=%s\n",value);

	bzero(value, 256);
	get_prop("PSK_KEY_TYPE", value, properties, prop_count);
	pos += snprintf(pos, end - pos, "PSK_KEY_TYPE=%s\n",value);
	free_prop(properties,prop_count);
	return pos-buf;
}

