/** @file ctrl_iface.c
 *   @brief This file contains functions for ctrl_iface
 *
 *  Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#ifndef CONFIG_NATIVE_WINDOWS
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/uio.h>
#endif /* CONFIG_NATIVE_WINDOWS */
#ifdef ANDROID
#include <cutils/sockets.h>
#endif
#include <grp.h>

#include "include/common.h"
#include "include/eloop.h"
#include "include/wapi.h"
#include "include/wapi_asue.h"
#include "include/wapi_asue_i.h"
#include "include/config.h"
#include "include/ctrl_iface.h"
#include "include/l2_packet.h"
#include "include/interface_inout.h"

/** Serial to string */
#define SERIAL2STR(a) (a)[0], (a)[1], (a)[2], (a)[3]
/** Serial string */
#define SERIALSTR "%02x%02x%02x%02x"

#define REPLY_MSG_MAX_LENGTH 1024
#define os_snprintf snprintf

//add by kaibing.chen for unix socket support
#ifndef WAPI_CTRL_INTERFACE
#define WAPI_CTRL_INTERFACE "DIR=/data/system/wpa_supplicant GROUP=system"
#endif /* WAPI_CTRL_INTERFACE */
#ifndef WAPI_CTRL_INTERFACE_PATH
#define WAPI_CTRL_INTERFACE_PATH "/data/system/wpa_supplicant/eth0"
#endif /* WAPI_CTRL_INTERFACE_PATH */

/**
 * @brief Get controller interface status
 *
 * @param wpa_s     Pointer to wpa_supplicant data
 * @param buf       Output buffer
 * @param buflen    Buffer length
 *
 * @return          Size of buffer written
 */
static int wpa_supplicant_ctrl_iface_status(struct wapi_asue_st *wpa_s,
					    char *buf, size_t buflen)
{
	char *pos;
	char MACbuf[256];
	int ret = 0;

	pos = buf;
	// get phone's MAC address 
	ret = os_snprintf(MACbuf, 256, MACSTR"\n" , MAC2STR(wpa_s->own_addr));
	if (ret < 0 || ret >= 256)
		return buflen;
	/* change wapi state machine */
       ret = lib_get_wapi_state();

       switch (ret)
       {
       	case 0:
			//pos[0] = WAPI_DISCONNECTED;
			strcpy(pos, "status=WAPI_SCANNING;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

       	case 1:
			//pos[0] = WAPI_ASSOCIATED;
			strcpy(pos, "status=WAPI_ASSOCIATED;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

       	case 2:
			//pos[0] = WAPI_CERT_HANDSHAKE;
			strcpy(pos, "status=WAPI_CERT_HANDSHAKE;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

       	case 3:
			//pos[0] = WAPI_3WAY_HANDSHAKE;
			strcpy(pos, "status=WAPI_3WAY_HANDSHAKE;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

       	case 4:
		case 5:
			//pos[0] = WAPI_GROUP_HANDSHAKE;
			strcpy(pos, "status=WAPI_GROUP_HANDSHAKE;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

       	case 6:
       	case 7:
			//pos[0] = WAPI_COMPLETED;
			strcpy(pos, "status=WAPI_COMPLETED;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

		case 8:
		case 9:
			//pos[0] = WAPI_3WAYING;
			strcpy(pos, "status=WAPI_3WAYING;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

		case 10:
			//pos[0] = WAPI_GROUPING;
			strcpy(pos, "status=WAPI_GROUPING;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

		default:
			//pos[0] = WAPI_DISCONNECTED;
			strcpy(pos, "status=WAPI_DISCONNECTED;");
			strncat(pos,MACbuf,strlen(MACbuf));
			break;

       }
	//pos[1] = 0;

	return strlen(pos);
}
 
static void wpa_supplicant_ctrl_iface_scan(
	struct wapi_asue_st *wpa_s, int sec, int usec)
{
	wpa_s->wapi_state = WAPI_INACTIVE;
	wpa_supplicant_req_scan( wpa_s, sec , usec );
}

static int wpa_supplicant_ctrl_iface_scan_results(
	struct wapi_asue_st *wpa_s, char *buf, size_t buflen)
{
	char *pos, *end, *retpos;
	struct wpa_scan_result *res;
	int i, ret;

	if (wpa_s->scan_results == NULL &&
	    wpa_supplicant_get_scan_results(wpa_s) < 0)
		return 0;
	if (wpa_s->scan_results == NULL)
		return 0;
	
	pos = buf;
	end = buf + buflen;
	ret = os_snprintf(pos, end - pos, "bssid / frequency / signal level / "
			  "flags / ssid\n");
	if (ret < 0 || ret >= end - pos)
		return pos - buf;
	pos += ret;

	for (i = 0; i < wpa_s->num_scan_results; i++) {
		retpos = pos;
		res = &wpa_s->scan_results[i];
		if(!res->wapi_ie_len){//filter none WAPI AP
			if( res->rsn_ie_len || res->wpa_ie_len || (res->caps & IEEE80211_CAP_IBSS ) )
				continue;
		}
		ret = os_snprintf(pos, end - pos, MACSTR "\t%d\t%d\t",
				  MAC2STR(res->bssid), res->freq, res->level);
		if (ret < 0 || ret >= end - pos)
			return retpos - buf;
		pos += ret;
		if (res->wapi_ie_len && res->caps & IEEE80211_CAP_PRIVACY) {
			if( res->wapi_ie[9] == 1 )//add by kaibing.chen
				ret = os_snprintf(pos, end - pos, "[WAPI-CERT]");
			if( res->wapi_ie[9] == 2 )
				ret = os_snprintf(pos, end - pos, "[WAPI-PSK]");
			if (ret < 0 || ret >= end - pos)
				return retpos - buf;
			pos += ret;
		}
		else {//add by kaibing.chen
			ret = os_snprintf(pos, end - pos, "[OPEN]");
			if (ret < 0 || ret >= end - pos)
				return retpos - buf;
			pos += ret;
		}
		ret = os_snprintf(pos, end - pos, "\t%s",
				  wpa_ssid_txt(res->ssid, res->ssid_len));
		if (ret < 0 || ret >= end - pos)
			return retpos - buf;
		pos += ret;

		ret = os_snprintf(pos, end - pos, "\n");
		if (ret < 0 || ret >= end - pos)
			return retpos - buf;
		pos += ret;
	}

	return pos - buf;
}


/**
 * @brief Check the contents of the certificate
 *
 * @param mesg      Certificate
 * @param msglen    Certificate length
 *
 * @return          0
 */
int Process_CGI_checkcert(u8 *mesg, int msglen)
{
	mesg = mesg;
	msglen = msglen;
	return 0;
}

/**
 * @brief Handle Message from CGI interface
 *
 * @param wpa_s     Pointer to wpa_supplicant data
 * @param recv_buf  Received buffer
 * @param readlen   Length of data read
 * @param cgicmd    Obtained CGI command
 *
 * @return          Bytes read
 */
int Process_CTRL_message(struct wapi_asue_st *wpa_s,u8 *recv_buf, int readlen, char * reply, int *cgicmd)
{
	int 		ret = 0;
	//u16	cmd = 0;
	char*	cmd = 0;

	//GETSHORT((recv_buf+2), cmd); 
	//printf(" in Process_CTRL_message, length is %d\n", readlen);
	recv_buf[readlen] = '\0';
	//copy the index in the first two bytes of recv_buf to reply message.
	memcpy(reply, recv_buf, 2);
	cmd = (char *)recv_buf + 2; 

	//printf(" in Process_CTRL_message, cmd is %s\n", cmd);
	do{
		//if (cmd == CGI_CMD_RELOAD)
       	if (strcmp(cmd, "RECONFIGURE") == 0)
		{
			ret = wpa_supplicant_reload_configuration(wpa_s);

			/*build reload response*/
			/*set type*/
            if (ret == 0) {
	            strcpy(reply+2, "OK\n");
                ret = 3 + 2;
            } else {
	            strcpy(reply+2, "FAIL\n");
                ret = 5 + 2;
            }
			//SETSHORT((recv_buf+2),CGI_RELOAD_RESPONSE);
			/*set length*/
			//SETSHORT((recv_buf+6),0);
			//ret = 8;
			usleep(2000);
			break;
		} 
       	else if (strcmp(cmd, "CHECKCERT") == 0)
		//else if (cmd == CGI_CMD_CHECKCERT)
		{
			ret = Process_CGI_checkcert(recv_buf, readlen);
			wpa_printf(MSG_DEBUG, "certificate verify result: %d\n", ret);   
			/*set type*/
			SETSHORT((recv_buf+2),CGI_CHECK_RESPONSE);
			/*set length*/
			SETSHORT((recv_buf+6),2);
			
			if (ret == 0)
				memset(recv_buf+8, 0x0000, 2);
			else
				memset(recv_buf+8, 0xFFFF, 2);
			ret = 10;
			break;
		}
       	else if (strcmp(cmd, "STATUS") == 0)
		//else if( cmd == CTRL_MSG_STATUS_REQ)
		{
			wpa_hexdump(MSG_DEBUG, "STATUS Req", recv_buf,readlen);   
			ret = wpa_supplicant_ctrl_iface_status(wpa_s, (char *)(reply+2), REPLY_MSG_MAX_LENGTH);
			/*set length*/
			//SETSHORT((recv_buf+6),2);
			ret = ret + 2;
			wpa_hexdump(MSG_DEBUG, "STATUS Res", reply,ret);   
			break;
		}
       	else if (strcmp(cmd, "SET_NETWORKS") == 0)
		{
			wpa_hexdump(MSG_DEBUG, "SET_NETWORKS Req", recv_buf,readlen);   
			//ret = wpa_supplicant_ctrl_iface_set_networks(wpa_s, (char *)(reply+2), REPLY_MSG_MAX_LENGTH);
			ret = ret + 2;
			wpa_hexdump(MSG_DEBUG, "SET_NETWORKS Res", reply,ret);   
			break;
		}
       	else if (strcmp(cmd, "GET_NETWORKS") == 0)
		{
			wpa_hexdump(MSG_DEBUG, "GET_NETWORKS Req", recv_buf,readlen);   
			//ret = wpa_supplicant_ctrl_iface_get_networks(wpa_s, (char *)(reply+2), REPLY_MSG_MAX_LENGTH);
			ret = ret + 2;
			wpa_hexdump(MSG_DEBUG, "GET_NETWORKS Res", reply,ret);   
			break;
		}
	else if (strcmp(cmd, "SCAN_RESULTS") == 0)
		{
			wpa_hexdump(MSG_DEBUG, "SCAN_RESULTS Req", recv_buf,readlen);   
			ret = wpa_supplicant_ctrl_iface_scan_results(wpa_s, (char *)(reply+2), REPLY_MSG_MAX_LENGTH);
			ret = ret + 2;
			wpa_hexdump(MSG_DEBUG, "SCAN_RESULTS Res", reply,ret);   
			break;

		}
	else if (strcmp(cmd, "SCAN") == 0)
		//add by kaibing.chen
		{
			wpa_hexdump(MSG_DEBUG, "SCAN Req", recv_buf,readlen);   
			wpa_supplicant_ctrl_iface_scan(wpa_s, 0 ,0 );
			wpa_hexdump(MSG_DEBUG, "SCAN Res", reply,ret);   
                        ret = 2 + 2;
			strcpy(reply+2,"ok");
			break;

		}
	}while(0);	
	*cgicmd = cmd;
	return ret;
}

/**
 * @brief Receive control interface message from CGI
 *
 * @param sock          Socket
 * @param eloop_ctx     Eloop context
 * @param sock_ctx      Socket context
 *
 * @return              None
 */
static void wpa_supplicant_ctrl_iface_receive_from_CGI(int sock, void *eloop_ctx,
					      void *sock_ctx)
{
	struct wapi_asue_st *wpa_s = eloop_ctx;
	u8  buf[2000];
	int res;
	//CTRL_IFACE_SOCK from;
	struct sockaddr_un from; //by kaibing.chen
	socklen_t fromlen = sizeof(from);
	int reply_len;
	int cmd = 0;
	const int reply_size = REPLY_MSG_MAX_LENGTH;
	char * reply = NULL;
	
	reply = malloc(reply_size);
	if (reply == NULL) {
		wpa_printf(MSG_ERROR, "in wpa_supplicant_ctrl_iface_receive_from_CGI, malloc fails");
		return;
	}
	res = recvfrom(sock, buf, sizeof(buf) - 1, 0,
		       (struct sockaddr *) &from, &fromlen);
	if (res < 0) {
		perror("recvfrom(ctrl_iface)");
		return;
	}
	wpa_hexdump(MSG_DEBUG, "From CGI",buf,  res);
	reply_len = Process_CTRL_message(wpa_s, buf, res, reply, &cmd);

	if (reply_len > 0) 
		sendto(sock, reply, reply_len, 0, (struct sockaddr *) &from, fromlen);

	free(reply);
}

/**
 * @brief Initialize control interface
 *
 * @param wpa_s Pointer to wpa_supplicant data
 *
 * @return      0 on success, -1 on failure
 *
 * Initialize the control interface and start receiving commands from external
 * programs.
 */
int wpa_supplicant_ctrl_iface_init_CGI(struct wapi_asue_st *wpa_s)
{/*
	CTRL_IFACE_SOCK addr;
	int s = -1;
#ifndef CONFIG_CTRL_IFACE_UDP
	char *fname = NULL;
#endif // CONFIG_CTRL_IFACE_UDP 

	wpa_s->cgi_sock = -1;

#ifdef CONFIG_CTRL_IFACE_UDP
	s = socket(PF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket(PF_INET)");
		goto fail;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr("127.0.0.1");
	addr.sin_port = htons(WPA_CGI_IFACE_PORT);
	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		perror("bind(AF_INET)");
		goto fail;
	}
#endif // CONFIG_CTRL_IFACE_UDP 

	wpa_s->cgi_sock = s;
	eloop_register_read_sock(s, wpa_supplicant_ctrl_iface_receive_from_CGI, wpa_s,
				 NULL);

	return 0;

fail:
	if (s >= 0)
		close(s);
#ifndef CONFIG_CTRL_IFACE_UDP
	if (fname) {
		unlink(fname);
		free(fname);
	}
#endif // CONFIG_CTRL_IFACE_UDP 
	return -1;
*/


	struct sockaddr_un addr;
	char *fname = NULL;
	gid_t gid = 0;
	int gid_set = 0;
	char *buf, *dir = NULL, *gid_str = NULL;
	struct group *grp;
	char *endp;

	wpa_s->cgi_sock	= -1;
	wpa_s->conf->ctrl_interface = strdup(WAPI_CTRL_INTERFACE);

	if (wpa_s->conf->ctrl_interface == NULL)
		goto fail;

	buf = strdup(wpa_s->conf->ctrl_interface);
	if (buf == NULL)
		goto fail;
#ifdef ANDROID
	snprintf(addr.sun_path, sizeof(addr.sun_path), "wpa_%s",
			wpa_s->conf->ctrl_interface);
	wpa_s->cgi_sock = android_get_control_socket(addr.sun_path);
	if (wpa_s->cgi_sock >= 0)
		goto havesock;
#endif
	if (strncmp(buf, "DIR=", 4) == 0) {
		dir = buf + 4;
		gid_str = strstr(dir, " GROUP=");
		if (gid_str) {
			*gid_str = '\0';
			gid_str += 7;
		}
	} else {
		dir = buf;
		gid_str = wpa_s->conf->ctrl_interface_group;
	}

	if (mkdir(dir, S_IRWXU | S_IRWXG) < 0) {
		if (errno == EEXIST) {
			wpa_printf(MSG_DEBUG, "Using existing control "
				   "interface directory.");
		} else {
			perror("mkdir[ctrl_interface]");
			goto fail;
		}
	}

	if (gid_str) {
		grp = getgrnam(gid_str);
		if (grp) {
			gid = grp->gr_gid;
			gid_set = 1;
			wpa_printf(MSG_DEBUG, "ctrl_interface_group=%d"
				   " (from group name '%s')",
				   (int) gid, gid_str);
		} else {
			// Group name not found - try to parse this as gid 
			gid = strtol(gid_str, &endp, 10);
			if (*gid_str == '\0' || *endp != '\0') {
				wpa_printf(MSG_DEBUG, "CTRL: Invalid group "
					   "'%s'", gid_str);
				goto fail;
			}
			gid_set = 1;
			wpa_printf(MSG_DEBUG, "ctrl_interface_group=%d",
				   (int) gid);
		}
	}

	if (gid_set && chown(dir, -1, gid) < 0) {
		perror("chown[ctrl_interface]");
		goto fail;
	}

	if (strlen(dir) + 1 + strlen(wpa_s->ifname) >=
	    sizeof(addr.sun_path))
		goto fail;

	wpa_s->cgi_sock = socket(PF_UNIX, SOCK_DGRAM, 0);
	if (wpa_s->cgi_sock < 0) {
		perror("socket(PF_UNIX)");
		goto fail;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_UNIX;
	fname = strdup (WAPI_CTRL_INTERFACE_PATH);
	if (fname == NULL)
		goto fail;
	strncpy(addr.sun_path, fname, sizeof(addr.sun_path));
	if (bind(wpa_s->cgi_sock, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		wpa_printf(MSG_DEBUG, "ctrl_iface bind(PF_UNIX) failed: %s",
			   strerror(errno));
		if (connect(wpa_s->cgi_sock, (struct sockaddr *) &addr,
			    sizeof(addr)) < 0) {
			wpa_printf(MSG_DEBUG, "ctrl_iface exists, but does not"
				   " allow connections - assuming it was left"
				   "over from forced program termination");
			if (unlink(fname) < 0) {
				perror("unlink[ctrl_iface]");
				wpa_printf(MSG_ERROR, "Could not unlink "
					   "existing ctrl_iface socket '%s'",
					   fname);
				goto fail;
			} else {
               wpa_printf(MSG_DEBUG, "in wpa_supplicant_ctrl_iface_init_CGI, unlink successfully");
            }
			if (bind(wpa_s->cgi_sock, (struct sockaddr *) &addr,
				 sizeof(addr)) < 0) {
				perror("bind(PF_UNIX)");
                wpa_printf(MSG_DEBUG, "in wpa_supplicant_ctrl_iface_init_CGI, after unlink, bind fails");
				goto fail;
			} else {
                wpa_printf(MSG_DEBUG, "in wpa_supplicant_ctrl_iface_init_CGI, bind successfully");
            }

			wpa_printf(MSG_DEBUG, "Successfully replaced leftover "
				   "ctrl_iface socket '%s'", fname);
		} else {
			wpa_printf(MSG_INFO, "ctrl_iface exists and seems to "
				   "be in use - cannot override it");
			wpa_printf(MSG_INFO, "Delete '%s' manually if it is "
				   "not used anymore", fname);
			free(fname);
			fname = NULL;
			goto fail;
		}
	}

	if (gid_set && chown(fname, -1, gid) < 0) {
		perror("chown[ctrl_interface/ifname]");
		goto fail;
	}

	if (chmod(fname, S_IRWXU | S_IRWXG) < 0) {
		perror("chmod[ctrl_interface/ifname]");
		goto fail;
	}
	free(fname);

havesock:
	eloop_register_read_sock(wpa_s->cgi_sock, wpa_supplicant_ctrl_iface_receive_from_CGI, wpa_s,
				 NULL);
	free(buf);
	
	return 0;

fail:
	if (wpa_s->cgi_sock >= 0)
		close(wpa_s->cgi_sock);
	if (fname) {
		unlink(fname);
		free(fname);
	}
	free(buf);
	
	return -1;

}

/**
 * @brief Send a control interface packet to monitors
 *
 * @param wpa_s    Pointer to wpa_supplicant data
 * @param level    Priority level of the message
 * @param buf      Message data
 * @param len      Message length
 *
 * @return         None
 *
 * Send a packet to all monitor programs attached to the control interface.
 */
 void wpa_supplicant_ctrl_iface_send(struct wapi_asue_st *wpa_s, int level,
					 char *buf, size_t len)
 {
 	return ;
 }
 
/**
 * @brief Deinitialize control interface
 *
 * @param wpa_s    Pointer to wpa_supplicant data
 *
 * @return         None
 *
 * Deinitialize the control interface that was initialized with
 * wpa_supplicant_ctrl_iface_init().
 */
 void wpa_supplicant_ctrl_iface_deinit(struct wapi_asue_st *wpa_s)
 {
		 if (wpa_s->ctrl_sock > -1) {
				 eloop_unregister_read_sock(wpa_s->ctrl_sock);
				 close(wpa_s->ctrl_sock);
				 wpa_s->ctrl_sock = -1;
		 }
 }

/**
 * @brief Deinitialize CGI interface
 *
 * @param wpa_s     Pointer to wpa_supplicant data
 *
 * @return          None
 */
 void wpa_supplicant_cgi_ctrl_iface_deinit(struct wapi_asue_st *wpa_s)
 {
		 if (wpa_s->cgi_sock > -1) {
				 eloop_unregister_read_sock(wpa_s->cgi_sock);
				 close(wpa_s->cgi_sock);
				 wpa_s->cgi_sock = -1;
		 }
 }
