/** @file wapi_asue.h
 *  @brief This header file contains data structures and function declarations of wapi_asue
 *
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 *  Copyright (C) 2008-2009, Marvell International Ltd. 
 */

#ifndef __WAPI_ASUE_H__
#define __WAPI_ASUE_H__

#include <sys/socket.h>
#include <netinet/in.h>

#include <sys/time.h>
#include "config.h"
#include "driver.h"
#include "wapi_config.h"
#include "l2_packet.h"
#include "wapi.h"

/** Ctrl Interface Socket */
#define CTRL_IFACE_SOCK struct sockaddr_in
/** Ctrl Interface Port */
#define WPA_CTRL_IFACE_PORT 9003
/** CGI Interface Port */
#define WPA_CGI_IFACE_PORT 9002
/** CGI Command Reload */
#define CGI_CMD_RELOAD		0x0212
/** CGI Reload Response */
#define CGI_RELOAD_RESPONSE	0x0213
/** CGI Command Check Cert */
#define CGI_CMD_CHECKCERT	0x0214
/** CGI Check Response */
#define CGI_CHECK_RESPONSE	0x0215

/** Enumaration:wapi_states */
typedef enum {
	WAPI_DISCONNECTED=0,		
	WAPI_INACTIVE,
	WAPI_SCANNING,
	WAPI_ASSOCIATING,
	WAPI_ASSOCIATED,
	WAPI_AUTHENICATING,
	WAPI_CERT_HANDSHAKE,
	WAPI_3WAY_HANDSHAKE,
	WAPI_3WAYING,
	WAPI_GROUP_HANDSHAKE,
	WAPI_GROUPING,
	WAPI_COMPLETED
} wapi_states;

/** Enumaration:wapi_wai_err */
typedef enum {
	WAI_ERR_INIT=0,
	WAI_ERR_BKID,
	WAI_ERR_CHALLENGE,
	WAI_ERR_IE,
	WAI_ERR_MIC,
} wapi_wai_err;

/** Enumaration:wpa_event_type */
typedef enum {
	EVENT_ASSOC, EVENT_DISASSOC, EVENT_MICHAEL_MIC_FAILURE,
	EVENT_SCAN_RESULTS, EVENT_ASSOCINFO, EVENT_INTERFACE_STATUS,
	EVENT_PMKID_CANDIDATE,EVENT_ASSOC_TIME
} wpa_event_type;

/** WAI header */
struct wai_hdr {
	/** Version */
	u8 version[2];
	/** Type */
	u8 type;
	/** Stype */
	u8 stype;
	/** Reserve */
	u8 reserve[2]; 
	/** Length */
	u16 length;
	/** Packet Sequence Number */
	u8 rxseq[2];	
	/** Fragment Sequence Number */
	u8 frag_sc;	
	/** Flag */
	u8 more_frag;				
	/* followed octets of data */
};// __attribute__ ((packed));

/** WAPI retry */
struct  wapi_retry{
	/** WAPI states */
	wapi_states states;
	/** WAI header */
	struct wai_hdr hdr;
	/** Retry count */
	u8 retry_count;
	/** Private */
	void *priv;
	/** Retry buffer */
	u8 *retry_buf;
	/** Length */
	u16  len;
	/** MAC */
	u8 mac[ETH_ALEN];
	/** Proto */
	u16 proto;
	/** CAllback function */
	int (*handle)(struct l2_packet_data *l2,  struct wapi_retry *retry);
};

/** WAPI state machine */
struct wapi_state_machine {
	/** WAPI retry */
	struct wapi_retry retry;
	/** number of entries in PMKSA cache */
	int bksa_count; 
	/** Own address */
	u8 own_addr[ETH_ALEN];
	/** Interface name */
	const char *ifname;
	/** BSSID */
	u8 bssid[ETH_ALEN];
	/** Pairwise cipher */
	unsigned int pairwise_cipher;
	/** Group cipher */
	unsigned int group_cipher;
	/** Key management */
	unsigned int key_mgmt;
	/**  Own WAPI/RSN IE from (Re)AssocReq */
	u8 assoc_wapi_ie[256];
	/** IE length */
	u8 assoc_wapi_ie_len;
	/** WAPI IE */
	u8 ap_wapi_ie[256];
	/** WAPI IE length */
	u8 ap_wapi_ie_len;
};

/** WAPI receive fragment */
struct wapi_rxfrag{
	/** Data */
	const u8 *data;
	/** Data length */
	int data_len;
	/** Maximum length */
	int maxlen;
	/** Rx fragment stamp */
	int rxfragstamp;
};

/** Byte Data */
typedef struct byte_data_{
	/** Length */
	u8 length;
	/** Data */
	u8 data[256];
}byte_data;

/** Structure retry buf */
struct _resendbuf_st{
	/** Count */
	u16 	cur_count;
	/** Length */
	u16		len;
	/** Data */
	void 	*data;
};

/** Typedef resendbuf_st */
typedef struct _resendbuf_st resendbuf_st;


/**
 * struct wapi_ssid - Network configuration data
 *
 * This structure includes all the configuration variables for a network. This
 * data is included in the per-interface configuration data as an element of
 * the network list, struct wpa_config::ssid. Each network block in the
 * configuration is mapped to a struct wpa_ssid instance.
 */
struct wapi_asue_st {
	/** L2 packet data */
	struct l2_packet_data *l2;
	/** Own address */
	u8 own_addr[ETH_ALEN];
	/** BSSID */
	u8 bssid[ETH_ALEN];
	/** Interface name */
	char ifname[100];
	/** Configure name */
	char *confname;
	/** WAPI config */
	struct wapi_config *conf;
	/** reassociation requested */
	int reassociate; /* reassociation requested */
	/** Disconnected flag */
	int disconnected; /* all connections disabled; i.e., do no reassociate
			   * before this has been cleared */
	/** WAI received */
	int wai_received;		   	
	/* Selected configuration (based on Beacon/ProbeResp WPA IE) */
	/** Pairwise cipher */
	int pairwise_cipher;
	/** Group cipher */
	int group_cipher;
	/** Key management */
	int key_mgmt;
	/** private data used by driver_ops */
	void *drv_priv;
	/** Scan results */
	struct wpa_scan_result *scan_results;
	/** Number of scan results */
	int num_scan_results;
	/** Driver */
	struct wpa_driver_ops *driver;
	/** WAPI state machine */
	struct wapi_state_machine *wapi_sm;
	/** UNIX domain socket for control interface */
	int ctrl_sock; /* UNIX domain socket for control interface or -1 if
			* not used */
	/** UNIX domain socket for control interface */
	int cgi_sock; /* UNIX domain socket for control interface or -1 if
			* not used */
	/** Rx fragment timestamp */
	int rxfragstamp;
	/** Rx fragment */
	struct wapi_rxfrag *rxfrag;
	/** Rx sequence */
	u16 rxseq;
	/** Tx sequence */
	u16 txseq;
	/** WAPI state */
	wapi_states wapi_state;
	/** WAI error */
	wapi_wai_err wai_err;
	/** Last WAI source */
	unsigned char last_wai_src[ETH_ALEN];
	/** Pad */
	u8 pad;
	/** Flag */
	u8 flag;
	/** Buffer */
	struct _resendbuf_st buf;
	/** Own WAPI/RSN IE from (Re)AssocReq */
	u8 assoc_wapi_ie[256]; 
	/** AP WAPI IE */
	u8 ap_wapi_ie[256];
	/** AP WAPI IE length */
	u8 ap_wapi_ie_len;
	/** assoc WAPI IE length */
	u8 assoc_wapi_ie_len;
};

int wpa_supplicant_reload_configuration(struct wapi_asue_st *wpa_s);
/**
 * wpa_supplicant_event - report a driver event for wpa_supplicant
 * @wpa_s: pointer to wpa_supplicant data; this is the @ctx variable registered
 *	with wpa_driver_events_init()
 * @event: event type (defined above)
 * @data: possible extra data for the event
 *
 * Driver wrapper code should call this function whenever an event is received
 * from the driver.
 */
void wpa_supplicant_event(struct wapi_asue_st *wpa_s, wpa_event_type event,
			  void *data);
struct wapi_asue_st * wapi_asue_alloc(void);
int wapi_asue_init(struct wapi_asue_st *wpa_s,
			       const char *confname, const char *driver,
			       const char *ifname);
const char * wpa_ssid_txt(u8 *ssid, size_t ssid_len);
const char * wpa_supplicant_state_txt(int state);
const char * wpa_mode_txt();
const char * wpa_signal_txt();
const char * wpa_wapi_txt(int wapi);
int wpa_supplicant_get_scan_results(struct wapi_asue_st *wpa_s);
int wpa_supplicant_init2(struct wapi_asue_st *wpa_s, int wait_for_interface);
void wpa_supplicant_terminate(int sig, void *eloop_ctx,
				     void *signal_ctx);
void wpa_supplicant_cancel_auth_timeout(struct wapi_asue_st *wpa_s);

void wpa_supplicant_deauthenticate(struct wapi_asue_st *wpa_s,
				   int reason_code);
void wpa_supplicant_deinit(struct wapi_asue_st *wpa_s);
int wpa_ether_send(struct wapi_asue_st *wpa_s, const u8 *dest, u16 proto,
		   const u8 *buf, size_t len);
void usage(void);
void license(void);
void print_version(void);

void wpa_msg(struct wapi_asue_st *wpa_s, int level, char *fmt, ...);
int wpa_sm_get_status(struct wapi_state_machine *sm, char *buf, size_t buflen);
void wpa_supplicant_req_scan(struct wapi_asue_st *wpa_s, int sec, int usec);
void wpa_supplicant_associate(struct wapi_asue_st *wpa_s,const u8 *ssid, int ssid_len);
void wpa_supplicant_disassociate(struct wapi_asue_st *wpa_s,	 int reason_code);
int wpa_config_set(struct wapi_asue_st *wpa_s,const char *var, const char *value, int line);

struct wapi_rxfrag *wpa_defrag(struct wapi_asue_st *wpa_s, struct wapi_rxfrag *rxbuf);
void reset_rebuf(struct wapi_asue_st *wpa_s);
void *free_rxfrag(struct wapi_rxfrag *frag);

#endif
