/** @file l2_packet_linux.c
  * @brief This file contains functions for l2_packet_linux
 *
 *  Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netpacket/packet.h>
#include <net/if.h>
#include <arpa/inet.h>

#include "common.h"
#include "eloop.h"
#include "l2_packet.h"

/** l2_packet_data structure */
struct l2_packet_data {
	/** packet socket for EAPOL frames */
	int fd;
	/** Interface name */
	char ifname[IFNAMSIZ + 1];
	/** Index */
	int ifindex;
	/** Address */
	u8 own_addr[ETH_ALEN];
	/** Callback function */
	void (*rx_callback)(void *ctx, const u8 *src_addr,
			    const u8 *buf, size_t len);
	/** Callback data */
	void *rx_callback_ctx;
	/** 1 = include layer 2 header, 0 = do not include header */
	int l2_hdr;
};

/** 
 *  @brief        Get own layer 2 address  
 *
 *  @param l2	 Pointer to internal l2_packet_data structure 
 *  @param addr	 Buffer for the own address
 *  @return      0 on success, -1 on failure
 */
int l2_packet_get_own_addr(struct l2_packet_data *l2, u8 *addr)
{
	memcpy(addr, l2->own_addr, ETH_ALEN);
	return 0;
}

/** 
 *  @brief       Send a packet     
 *
 *  @param l2	 Pointer to internal l2_packet_data structure 
 *  @param dst_addr	 Destination address for the packet
 *  @param proto     Protocol/ethertype for the packet in host byte order
 *  @param buf       Packet contents to be sent
 *  @param len 	     Length of the buffer
 *  @return          >=0 on success, <0 on failure
 */
int l2_packet_send(struct l2_packet_data *l2, const u8 *dst_addr, u16 proto,
		   const u8 *buf, size_t len)
{
	u8 txbuf[2312]= {0,};
	u8 *pos = txbuf;
	int ret;

	if (l2 == NULL){
		return -1;
	} else {
		struct sockaddr_ll ll;
		memset(&ll, 0, sizeof(ll));
		ll.sll_family = AF_PACKET;
		ll.sll_ifindex = l2->ifindex;
		ll.sll_protocol = htons(proto);
		ll.sll_halen = ETH_ALEN;

		memcpy (pos, dst_addr,   ETH_ALEN); pos += ETH_ALEN;
		memcpy (pos, l2->own_addr,   ETH_ALEN);pos += ETH_ALEN;
		memcpy(pos,  &ll.sll_protocol, 2); pos+=2;
		memcpy (pos, buf, len);
		wpa_hexdump(MSG_DEBUG, "txbuf", txbuf, len + ETH_HLEN);
		ret = sendto(l2->fd, txbuf, len+ETH_HLEN, 0, (struct sockaddr *) &ll,
			     sizeof(ll));
		wpa_printf(MSG_DEBUG, "l2_packet_send: ret = '%d'", ret);
		if (ret < 0)
			perror("l2_packet_send - sendto");
	}
	return ret;
}

/** 
 *  @brief       Receive a packet     
 *
 *  @param sock       Socket
 *  @param eloop_ctx  Pointer to l2_packet_data structure 
 *  @param sock_ctx   Context data
 *  @return           None
 */
static void l2_packet_receive(int sock, void *eloop_ctx, void *sock_ctx)
{
	struct l2_packet_data *l2 = eloop_ctx;
	u8 buf[2300];
	int res;
	struct l2_ethhdr l2_hdr;

	sock_ctx = sock_ctx;
	
	memset(&l2_hdr, 0, sizeof(l2_hdr));

	res = recvfrom(sock, buf, sizeof(buf), 0, NULL,NULL);

	if (res < 0) {
		perror("l2_packet_receive - recvfrom");
		return;
	}
        memcpy(&l2_hdr, buf, ETH_HLEN);
        l2_hdr.h_proto = ntohs (l2_hdr.h_proto);
        res -= ETH_HLEN;
        if (res > 0) memmove(buf, (char *)buf+ETH_HLEN, res);
        else res = 0;

	l2->rx_callback(l2->rx_callback_ctx, l2_hdr.h_source, buf, res);
}

/** 
 *  @brief       Initialize l2_packet interface     
 *
 *  @param ifname          Interface name      
 *  @param own_addr        Optional own MAC address
 *  @param protocol        Ethernet protocol number in host byte order
 *  @param rx_callback     Callback function that will be called for each received packet
 *  @param rx_callback_ctx Callback data (ctx) for calls to rx_callback()
 *  @param l2_hdr          1 = include layer 2 header, 0 = do not include header
 *  @return                Pointer to internal data or NULL on failure
 */
struct l2_packet_data * l2_packet_init(
	const char *ifname, 
	const u8 *own_addr, 
	unsigned short protocol,
	void (*rx_callback)(void *ctx, const u8 *src_addr,
			    const u8 *buf, size_t len),
	void *rx_callback_ctx,
	int l2_hdr)
{
	struct l2_packet_data *l2;
	struct ifreq ifr;
	struct sockaddr_ll ll;

	own_addr = own_addr;
	
	l2 = malloc(sizeof(struct l2_packet_data));
	if (l2 == NULL)
		return NULL;
	memset(l2, 0, sizeof(*l2));
	strncpy(l2->ifname, ifname, sizeof(l2->ifname));
	l2->rx_callback = rx_callback;
	l2->rx_callback_ctx = rx_callback_ctx;
	l2->l2_hdr = l2_hdr;

	//l2->fd = socket(PF_PACKET, l2_hdr ? SOCK_RAW : SOCK_DGRAM,
	//		htons(protocol));
	l2->fd = socket(PF_PACKET,  SOCK_RAW ,	htons(protocol));
	if (l2->fd < 0) {
		perror("socket(PF_PACKET)");
		free(l2);
		return NULL;
	}
	strncpy(ifr.ifr_name, l2->ifname, sizeof(ifr.ifr_name));
	if (ioctl(l2->fd, SIOCGIFINDEX, &ifr) < 0) {
		perror("ioctl[SIOCGIFINDEX]");
		close(l2->fd);
		free(l2);
		return NULL;
	}
	l2->ifindex = ifr.ifr_ifindex;

	memset(&ll, 0, sizeof(ll));
	ll.sll_family = PF_PACKET;
	ll.sll_ifindex = ifr.ifr_ifindex;
	ll.sll_protocol = htons(protocol);
	if (ioctl(l2->fd, SIOCGIFHWADDR, &ifr) < 0) {
		perror("ioctl[SIOCGIFHWADDR]");
		close(l2->fd);
		free(l2);
		return NULL;
	}
	memcpy(l2->own_addr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

	eloop_register_read_sock(l2->fd, l2_packet_receive, l2, NULL);

	return l2;
}

/** 
 *  @brief       Deinitialize l2_packet interface     
 *
 *  @param l2    Pointer to internal l2_packet data
 *  @return      None
 */
void l2_packet_deinit(struct l2_packet_data *l2)
{
	if (l2 == NULL)
		return;

	if (l2->fd >= 0) {
		eloop_unregister_read_sock(l2->fd);
		close(l2->fd);
	}
		
	free(l2);
}

/** 
 *  @brief       Get the current IP address from the interface     
 *
 *  @param l2    Pointer to internal l2_packet data
 *  @param buf   Buffer for the IP address
 *  @param len   Maximum length of the buf
 *  @return      0 on success, -1 on failure
 */
int l2_packet_get_ip_addr(struct l2_packet_data *l2, char *buf, size_t len)
{
	int s;
	struct ifreq ifr;
	struct sockaddr_in *saddr;

	s = socket(PF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		return -1;
	}
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, l2->ifname, sizeof(ifr.ifr_name));
	if (ioctl(s, SIOCGIFADDR, &ifr) < 0) {
		perror("ioctl[SIOCGIFADDR]");
		close(s);
		return -1;
	}
	close(s);
	saddr = (struct sockaddr_in *) &ifr.ifr_addr;
	if (saddr->sin_family != AF_INET)
		return -1;
	snprintf(buf, len, "%s", inet_ntoa(saddr->sin_addr));
	return 0;
}

/** 
 *  @brief       Notify l2_packet about start of authentication     
 *
 *  @param l2    Pointer to internal l2_packet data
 *  @return      None
 */
void l2_packet_notify_auth_start(struct l2_packet_data *l2)
{
	l2 = l2;
}

