/** @file main.c
  * @brief This file contains functions for main
 *
 *  Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>

#include "include/common.h"
#include "include/wapi_asue.h"
#include "include/eloop.h"
#include "include/interface_inout.h"
extern int wpa_debug_level ;
extern int wpa_debug_show_keys;
extern int wpa_debug_timestamp;

/** 
 *  @brief       Reconfigure wpa_supplicant     
 *
 *  @param sig        Signal number    
 *  @param eloop_ctx  eloop context    
 *  @param signal_ctx signal context    
 *  @return      None
 */
static void wpa_supplicant_reconfig(int sig, void *eloop_ctx,
				    void *signal_ctx)
{
	struct wapi_asue_st *wpa_s = eloop_ctx;

	signal_ctx = signal_ctx;	
	wpa_printf(MSG_DEBUG, "Signal %d received - reconfiguring", sig);
	if (wpa_supplicant_reload_configuration(wpa_s) < 0) {
		eloop_terminate();
	}
}

/** 
 *  @brief       Workaround   
 *
 *  @return      None
 */
static void wapi_supplicant_fd_workaround(void)
{
	int s, i;
	/* When started from pcmcia-cs scripts, wpa_supplicant might start with
	 * fd 0, 1, and 2 closed. This will cause some issues because many
	 * places in wpa_supplicant are still printing out to stdout. As a
	 * workaround, make sure that fd's 0, 1, and 2 are not used for other
	 * sockets. */
	for (i = 0; i < 3; i++) {
		s = open("/dev/null", O_RDWR);
		if (s > 2) {
			close(s);
			break;
		}
	}
}

/** 
 *  @brief          The main function 
 *
 *  @param argc     Argument count
 *  @param argv     Pointer to the argument list
 *  @return         0 on success, -1 on failure 
 */
int main(int argc, char *argv[])
{
	struct wapi_asue_st *head, *wpa_s;
	int c;
	const char *confname, *driver, *ifname;
	int daemonize = 0, wait_for_interface = 0,  exitcode;

	head = wpa_s = wapi_asue_alloc();
	if (wpa_s == NULL)
		return -1;

	wapi_supplicant_fd_workaround();
	eloop_init(head);

	ifname = confname = driver = NULL;

	for (;;) {
		c = getopt(argc, argv, "Bc:D:dehi:KLNqtvw");
		if (c < 0)
			break;
		switch (c) {
		case 'B':
			daemonize++;
			break;
		case 'c':
			confname = optarg;
			break;
		case 'd':
			wpa_debug_level--;
			break;
		case 'h':
			usage();
			return -1;
		case 'i':
			ifname = optarg;
			break;
		case 'K':
			wpa_debug_show_keys++;
			break;
		case 'q':
			wpa_debug_level++;
			break;
		case 't':
			wpa_debug_timestamp++;
			break;
		case 'v':
			print_version();
			return -1;
		case 'w':
			wait_for_interface++;
			break;
		default:
			usage();
			return -1;
		}
	}
	

	if (WIFI_lib_init())
	{
		exitcode = -1;
		goto cleanup;
	}

	if (wapi_asue_init(wpa_s, confname, driver, ifname))
		return -1;
	if (wpa_supplicant_init2(wpa_s,  wait_for_interface)) {
		exitcode = -1;
		goto cleanup;
	}
	
	exitcode = 0;

	if (wait_for_interface && daemonize) {
		wpa_printf(MSG_DEBUG, "Daemonize..");
		if (daemon(0, 0)) {
			perror("daemon");
			exitcode = -1;
			goto cleanup;
		}
	}
	eloop_register_signal(SIGINT, wpa_supplicant_terminate, NULL);
	eloop_register_signal(SIGTERM, wpa_supplicant_terminate, NULL);
#ifndef CONFIG_NATIVE_WINDOWS
	eloop_register_signal(SIGHUP, wpa_supplicant_reconfig, NULL);
#endif /* CONFIG_NATIVE_WINDOWS */

	eloop_run();

	wpa_supplicant_deauthenticate(wpa_s, REASON_DEAUTH_LEAVING);

cleanup:

	if (wpa_s) {
		wpa_supplicant_deinit(wpa_s);
		free(wpa_s);
	}
	
	WIFI_lib_exit();

	eloop_destroy();

#ifdef CONFIG_NATIVE_WINDOWS
	WSACleanup();
#endif /* CONFIG_NATIVE_WINDOWS */

	return exitcode;
}

