/** @file wapi_asue.c
  * @brief This file contains functions for wapi_asue
 *
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 *  Copyright (C) 2008-2009, Marvell International Ltd. 
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
/* for declaration of function `inet_addr'*/
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/time.h>

#include "include/wireless_copy.h"
#include "include/common.h"
#include "include/version.h"
#include "include/wapi_asue.h"
#include "include/wapi.h"
#include "include/ctrl_iface.h"
#include "include/wapi_config.h"
#include "include/wapi_sm.h"
#include "include/wapi_asue_i.h"
#include "include/eloop.h"
#include "include/wapi_asue_retry.h"
#include "include/driver_wext.h"
#include "include/interface_inout.h"

static void wpa_supplicant_cleanup(struct wapi_asue_st *wpa_s);
static int wpa_supplicant_set_driver(struct wapi_asue_st *wpa_s,
				     const char *name);
static int wpa_supplicant_driver_init(struct wapi_asue_st *wpa_s,
				      int wait_for_interface);
void wpa_supplicant_req_scan(struct wapi_asue_st *wpa_s, int sec, int usec);

extern struct wpa_driver_ops *wpa_supplicant_drivers[];

static const char *wpa_supplicant_version = "wapi_ASUE v" VERSION_STR "\n"
"Copyright (C) 2006-2008, Iwncomm Ltd.\nCopyright (C) 2008-2009, Marvell International Ltd. ";

/** 
 *  @brief      Display wapic usage 
 */
void usage()
{
	int i;
	printf("%s\n\n"
	       "usage:\n"
	       "  %s [-BddehqqvwW] [-P<pid file>] "
	       "[-g<global ctrl>] \\\n"
	       "        -i<ifname> -c<config file> [-C<ctrl>]"
	       "\n"
	       "drivers:\n",
	       wpa_supplicant_version,  "wapic");

	for (i = 0; wpa_supplicant_drivers[i]; i++) {
		printf("  %s = %s\n",
		       wpa_supplicant_drivers[i]->name,
		       wpa_supplicant_drivers[i]->desc);
	}

#ifndef CONFIG_NO_STDOUT_DEBUG
	printf("options:\n"
	       "  -B = run daemon in the background\n"
	       "  -c = Configuration file\n"
	       "  -C = ctrl_interface parameter (only used if -c is not)\n"
	       "  -i = interface name\n"
	       "  -d = increase debugging verbosity (-dd even more)\n"
	       //"  -g = global ctrl_interface\n"
	       "  -K = include keys (passwords, etc.) in debug output\n"
	       "  -t = include timestamp in debug messages\n"
	       "  -h = show this help text\n"
	       //"  -L = show license (GPL and BSD)\n"
	       "  -p = driver parameters\n"
	       "  -P = PID file\n"
	       "  -q = decrease debugging verbosity (-qq even less)\n"
	       "  -v = show version\n"
	       "  -w = wait for interface to be added, if needed\n"
	       "  -W = wait for a control interface monitor before starting\n");

	printf("example:\n"
	       "  wapic -i wlan0 -c /etc/wapi.conf\n");
#endif /* CONFIG_NO_STDOUT_DEBUG */
}

/** 
 *  @brief   Print version 
 */
void print_version(void)
{
	printf("%s\n", wpa_supplicant_version);
}

/** 
 *  @brief     Allocate wapi_asue_st structure  
 *
 *  @return    A pointer to wapi_asue_st structure on success,
 *             otherwise NULL
 */
struct wapi_asue_st * wapi_asue_alloc(void)
{
	struct wapi_asue_st *wpa_s;

	wpa_s = malloc(sizeof(*wpa_s));
	if (wpa_s == NULL)
		return NULL;
	memset(wpa_s, 0, sizeof(*wpa_s));
#ifdef CONFIG_XSUPPLICANT_IFACE
	wpa_s->dot1x_s = -1;
#endif /* CONFIG_XSUPPLICANT_IFACE */

	return wpa_s;
}

/** 
 *  @brief        Cleanup the wpa_supplicant  
 *
 *  @param wpa_s  A Pointer to wpa_asue_st structure
 *  @return       None
 */
static void wpa_supplicant_cleanup(struct wapi_asue_st *wpa_s)
{
	l2_packet_deinit(wpa_s->l2);
	wpa_s->l2 = NULL;

	//wpa_supplicant_ctrl_iface_deinit(wpa_s);
	wpa_supplicant_cgi_ctrl_iface_deinit(wpa_s);
	if (wpa_s->conf != NULL) {
		wpa_config_free(wpa_s->conf);
		wpa_s->conf = NULL;
	}

	free(wpa_s->confname);
	wpa_s->confname = NULL;

	if(wpa_s->scan_results)
	{
		free(wpa_s->scan_results);
		wpa_s->scan_results = NULL;
		wpa_s->num_scan_results = 0;
	}
	reset_rebuf(wpa_s);
	wpa_s->rxfrag = free_rxfrag(wpa_s->rxfrag);
}

/** 
 *  @brief        Set driver  
 *
 *  @param wpa_s  A Pointer to wpa_asue_st structure
 *  @param name   driver name
 *  @return       0 on success, -1 on failure
 */
static int wpa_supplicant_set_driver(struct wapi_asue_st *wpa_s,
				     const char *name)
{
	int i;

	if (wpa_s == NULL)
		return -1;

	if (wpa_supplicant_drivers[0] == NULL) {
		wpa_printf(MSG_ERROR, "No driver interfaces build into "
			   "wpa_supplicant.");
		return -1;
	}

	if (name == NULL) {
		/* default to first driver in the list */
		wpa_s->driver = wpa_supplicant_drivers[0];
		return 0;
	}

	for (i = 0; wpa_supplicant_drivers[i]; i++) {
		if (strcmp(name, wpa_supplicant_drivers[i]->name) == 0) {
			wpa_s->driver = wpa_supplicant_drivers[i];
			return 0;
		}
	}

	printf("Unsupported driver '%s'.\n", name);
	return -1;
}

/** 
 *  @brief        Deinitialize wpa_supplicant  
 *
 *  @param wpa_s  A Pointer to wpa_asue_st structure
 *  @return       None
 */
void wpa_supplicant_deinit(struct wapi_asue_st *wpa_s)
{
	if (wpa_s->drv_priv) {
		int ret = -1;
		ret = wpa_drv_set_wpa(wpa_s, 0);
		if (ret < 0) {
			fprintf(stderr, "Failed to disable WPA in the "
				"driver.\n");
		}
		wpa_drv_deinit(wpa_s);
	}
	wpa_supplicant_cleanup(wpa_s);
}

/** 
 *  @brief        Initialize   
 *
 *  @param wpa_s  A Pointer to wpa_asue_st structure
 *  @param confname Configuration file name
 *  @param driver Driver
 *  @param ifname Interface name
 *  @return       0 on success, otherwise -1
 */
int wapi_asue_init(struct wapi_asue_st *wpa_s,
			       const char *confname, const char *driver,
			       const char *ifname)
{
	wpa_printf(MSG_DEBUG, "Initializing interface '%s' conf '%s' driver "
		   "'%s'", ifname, confname, driver ? driver : "default");

	if (wpa_supplicant_set_driver(wpa_s, driver) < 0) {
		return -1;
	}

	if (ifname == NULL) {
		usage();
		printf("\nInterface name is required.\n");
		return -1;
	}
	if (strlen(ifname) >= sizeof(wpa_s->ifname)) {
		printf("Too long interface name '%s'.\n", ifname);
		return -1;
	}
	strncpy(wpa_s->ifname, ifname, sizeof(wpa_s->ifname));

	wpa_s->drv_priv = wpa_drv_init(wpa_s, wpa_s->ifname);
	if (wpa_s->drv_priv == NULL) {
		fprintf(stderr, "Failed to initialize driver interface\n");
		return -1;
	}

	if (confname) {
		wpa_s->confname = rel2abs_path(confname);
		if (wpa_s->confname == NULL) {
			wpa_printf(MSG_ERROR, "Failed to get absolute path "
				   "for configuration file '%s'.", confname);
			return -1;
		}
		wpa_printf(MSG_DEBUG, "Configuration file '%s' -> '%s'",
			   confname, wpa_s->confname);
		wpa_s->conf = wpa_config_read(wpa_s->confname);
		if (wpa_s->conf == NULL) {
			printf("Failed to read configuration file '%s'.\n",
			       wpa_s->confname);
			return -1;
		}
	}

	if (wpa_s->conf == NULL || wpa_s->conf->ssid == NULL) {
		usage();
		printf("\nNo networks (SSID) configured.\n");
		return -1;
	}
	return 0;
}

/** 
 *  @brief        Initialize interface  
 *
 *  @param wpa_s  A Pointer to wpa_asue_st structure
 *  @param wait_for_interface Flag: whether it is wait for interface or not 
 *  @return       0 on success, otherwise -1
 */
int wpa_supplicant_init2(struct wapi_asue_st *wpa_s, int wait_for_interface)
{
	const char *ifname;
	

	/* RSNA Supplicant Key Management - INITIALIZE */

	/* Initialize driver interface and register driver event handler before
	 * L2 receive handler so that association events are processed before
	 * EAPOL-Key packets if both become available for the same select()
	 * call. */

	ifname = wpa_drv_get_ifname(wpa_s);
	if (ifname && strcmp(ifname, wpa_s->ifname) != 0) {
		wpa_printf(MSG_DEBUG, "Driver interface replaced interface "
			   "name with '%s'", ifname);
		strncpy(wpa_s->ifname, ifname, sizeof(wpa_s->ifname));
	}

	if (wpa_supplicant_driver_init(wpa_s, wait_for_interface) < 0) {
		return -1;
	}
	
	if (wpa_supplicant_ctrl_iface_init_CGI(wpa_s)) {
		printf("Failed to initialize control interface UI.\n"
		       "You may have another wpa_supplicant process already "
		       "running or the file was\n"
		       "left by an unclean termination of wpa_supplicant in "
		       "which case you will need\n"
		       "to manually remove this file before starting "
		       "wpa_supplicant again.\n");
	}

	return 0;
}

/** 
 *  @brief        Reload configuration  
 *
 *  @param wpa_s  A Pointer to wpa_asue_st structure
 *  @return       0 on success, otherwise -1
 */
int wpa_supplicant_reload_configuration(struct wapi_asue_st *wpa_s)
{
	struct wapi_config *conf;

	if (wpa_s->confname == NULL)
		return -1;
	conf = wpa_config_read(wpa_s->confname);
	if (conf == NULL) {
		wpa_msg(wpa_s, MSG_ERROR, "Failed to parse the configuration "
			"file '%s' - exiting", wpa_s->confname);
		return -1;
	}

	wpa_config_free(wpa_s->conf);
	wpa_s->conf = conf;
	wpa_s->pairwise_cipher = wpa_s->conf->pairwise_cipher;
	wpa_s->group_cipher = wpa_s->conf->group_cipher;
	//wpa_s->reassociate = 1;
	wpa_msg(wpa_s, MSG_DEBUG, "Reconfiguration completed");
	
//	wapi_usksa_init(&wpa_s->wapi_sm->usksa);
//	wapi_msksa_init(&wpa_s->wapi_sm->msksa);
//	wapi_bksa_init(&wpa_s->wapi_sm->bksa, wpa_s->conf->psk_bk);
//	wpa_asue_initialize_wie(wpa_s);
//	wpa_supplicant_req_scan(wpa_s, 0, 0);
	if (wpa_drv_set_wpa(wpa_s, wpa_s->conf->wapi_policy) < 0) {
		fprintf(stderr, "Failed to enable WPA in the driver.\n");
		return -1;
	}

	return 0;
}

/** 
 *  @brief        Terminate wpa_supplicant  
 *
 *  @param sig        Signal number
 *  @param eloop_ctx  eloop context
 *  @param signal_ctx Signal context
 *  @return           None
 */
void wpa_supplicant_terminate(int sig, void *eloop_ctx,
				     void *signal_ctx)
{
	struct wapi_asue_st *wpa_s = eloop_ctx;
	signal_ctx =signal_ctx;
	
	wpa_msg(wpa_s, MSG_INFO, "Signal %d received - terminating",
			sig);
	eloop_terminate();
}

/** 
 *  @brief        Conditional printf  
 *
 *  @param wpa_s  A pointer to a wapi_asue_st structure
 *  @param level  severity level of the message
 *  @param fmt    printf format string, followed by optional arguments
 *  @return       None
 */
void wpa_msg(struct wapi_asue_st *wpa_s, int level, char *fmt, ...)
{
	va_list ap;
	char *buf;
	const int buflen = 2048;
	int len;

	buf = malloc(buflen);
	if (buf == NULL) {
		printf("Failed to allocate message buffer for:\n");
		va_start(ap, fmt);
		vprintf(fmt, ap);
		printf("\n");
		va_end(ap);
		return;
	}
	va_start(ap, fmt);
	len = vsnprintf(buf, buflen, fmt, ap);
	va_end(ap);
	wpa_printf(level, "%s", buf);
	free(buf);
}

/** 
 *  @brief             Scan for networks  
 *
 *  @param eloop_ctx   A pointer to a wapi_asue_st structure
 *  @param timeout_ctx Timeout context
 *  @return            None
 */
static void wpa_supplicant_scan(void *eloop_ctx, void *timeout_ctx)
{
	struct wapi_asue_st *wpa_s = eloop_ctx;
	u8 *ssid = NULL;
	int enabled;
	
	
	if (wpa_s->disconnected)
		return;

	enabled = 0;
	ssid = wpa_s->conf->ssid;
	if (wpa_s->conf->disabled) 
	{
		wpa_printf(MSG_DEBUG, "No enabled networks - do not scan");
		wpa_s->wapi_state = WAPI_INACTIVE;
		return;
	}

	if (wpa_s->wapi_state == WAPI_DISCONNECTED ||
	    wpa_s->wapi_state == WAPI_INACTIVE)
		wpa_s->wapi_state = WAPI_SCANNING;

        ssid = NULL;// for broadcast ,not specific
	wpa_printf(MSG_DEBUG, "Starting AP scan (%s SSID)",
		   ssid ? "specific": "broadcast");

	if (wpa_drv_scan(wpa_s, ssid ? ssid : NULL,
			 ssid ? wpa_s->conf->ssid_len : 0)) {
		wpa_printf(MSG_WARNING, "Failed to initiate AP scan.");
		wpa_supplicant_req_scan(wpa_s, 10, 0);
	}
}

/** 
 *  @brief        Setting scan request  
 *
 *  @param wpa_s  A pointer to a wapi_asue_st structure
 *  @param sec    Number of seconds
 *  @param usec   Number of microseconds
 *  @return       None
 */
void wpa_supplicant_req_scan(struct wapi_asue_st *wpa_s, int sec, int usec)
{
	wpa_msg(wpa_s, MSG_DEBUG, "Setting scan request: %d sec %d usec",
		sec, usec);
	eloop_cancel_timeout(wpa_supplicant_scan, wpa_s, NULL);
	eloop_register_timeout(sec, usec, wpa_supplicant_scan, wpa_s, NULL);
}

/** 
 *  @brief        Get scan results  
 *
 *  @param wpa_s  A pointer to a wapi_asue_st structure
 *  @return       0 on success, otherwise -1
 */
int wpa_supplicant_get_scan_results(struct wapi_asue_st *wpa_s)
{
#define SCAN_AP_LIMIT 128
	struct wpa_scan_result *results, *tmp;
	int num;

	results = malloc(SCAN_AP_LIMIT * sizeof(struct wpa_scan_result));
	if (results == NULL) {
		wpa_msg(wpa_s,MSG_WARNING, "Failed to allocate memory for scan "
			   "results");
		return -1;
	}

	num = wpa_drv_get_scan_results(wpa_s, results, SCAN_AP_LIMIT);
	wpa_msg(wpa_s,MSG_DEBUG, "Scan results: %d", num);
	if (num < 0) {
		wpa_printf(MSG_DEBUG, "Failed to get scan results");
		free(results);
		return -1;
	}
	if (num > SCAN_AP_LIMIT) {
		wpa_msg(wpa_s,MSG_INFO, "Not enough room for all APs (%d < %d)",
			   num, SCAN_AP_LIMIT);
		num = SCAN_AP_LIMIT;
	}

	/* Free unneeded memory for unused scan result entries */
	tmp = realloc(results, num * sizeof(struct wpa_scan_result));
	if (tmp || num == 0) {
		results = tmp;
	}

	free(wpa_s->scan_results);
	wpa_s->scan_results = results;
	wpa_s->num_scan_results = num;

	return 0;
}

/** 
 *  @brief        Deauthenticate 
 *
 *  @param wpa_s        A pointer to a wapi_asue_st structure
 *  @param reason_code  Reason code
 *  @return             None
 */
void wpa_supplicant_deauthenticate(struct wapi_asue_st *wpa_s,
				   int reason_code)
{
	u8 *addr = NULL;
	wpa_s->wapi_state = WAPI_DISCONNECTED;
	if (memcmp(wpa_s->bssid, "\x00\x00\x00\x00\x00\x00", ETH_ALEN) != 0) {
		wpa_drv_deauthenticate(wpa_s, wpa_s->bssid, reason_code);
		addr = wpa_s->bssid;
	}
}

/** 
 *  @brief            Not implemented yet 
 *
 *  @param wpa_s      A pointer to a wapi_asue_st structure
 *  @param src_addr   Address
 *  @param buf        Buffer
 *  @param len        Length of the buffer
 *  @return           None
 */
void wapi_sm_rx_wai(struct wapi_asue_st *wpa_s,
			    const unsigned char *src_addr, const unsigned char *buf,
			    size_t len);

/** 
 *  @brief            WAI rx packet indication
 *
 *  @param ctx        context
 *  @param src_addr   Address
 *  @param buf        Buffer
 *  @param len        Length of the buffer
 *  @return           None
 */
static void wpa_supplicant_rx_wai(void *ctx, const u8 *src_addr, const u8 *buf, size_t len)
{
	
	WAI_RX_packets_indication(buf, len);
}

/** 
 *  @brief            Initialize driver 
 *
 *  @param wpa_s              A pointer to a wapi_asue_st structure
 *  @param wait_for_interface Flag: wheather wait for interface or not
 *  @return                   0 on success, otherwise -1
 */
static int wpa_supplicant_driver_init(struct wapi_asue_st *wpa_s,
				      int wait_for_interface)
{
	for (;;) {
		wpa_s->l2 = l2_packet_init(
					   wpa_s->ifname,
					   wpa_drv_get_mac_addr(wpa_s),
					   ETH_P_WAI,
					   wpa_supplicant_rx_wai, 
					   wpa_s,
					   0);
		if (wpa_s->l2)
			break;
		else if (!wait_for_interface)
			return -1;
		printf("Waiting for interface..\n");
		sleep(5);
	}

	if (l2_packet_get_own_addr(wpa_s->l2, wpa_s->own_addr)) {
		fprintf(stderr, "Failed to get own L2 address\n");
		return -1;
	}

	wpa_printf(MSG_DEBUG, "Own MAC address: " MACSTR,
		   MAC2STR(wpa_s->own_addr));

	if (wpa_drv_set_wpa(wpa_s, wpa_s->conf->wapi_policy) < 0) {
		fprintf(stderr, "Failed to enable WPA in the driver.\n");
		return -1;
	}
	wpa_supplicant_req_scan(wpa_s,0, 100000);
	return 0;
}

/** 
 *  @brief          Convert SSID to a printable string 
 *
 *  @param ssid     SSID
 *  @param ssid_len SSID length
 *  @return         Pointer to a printable string
 */
const char * wpa_ssid_txt(u8 *ssid, size_t ssid_len)
{
	static char ssid_txt[MAX_SSID_LEN + 1];
	char *pos;

	if (ssid_len > MAX_SSID_LEN)
		ssid_len = MAX_SSID_LEN;
	memcpy(ssid_txt, ssid, ssid_len);
	ssid_txt[ssid_len] = '\0';
	for (pos = ssid_txt; *pos != '\0'; pos++) {
		if ((u8) *pos < 32 || (u8) *pos >= 127)
			*pos = '_';
	}
	return ssid_txt;
}

/** 
 *  @brief          Convert wapi to a printable string 
 *
 *  @param wapi     wapi
 *  @return         Pointer to a printable string
 */
const char * wpa_wapi_txt(int wapi)
{
	switch (wapi) {
	case 0:
		return "0";//"Authentication Succeed;Data been encrypted."
	case 1:
		return "1";//"Certificate is being authenticated...."
	case 2:
		return "2";//"Unicast Key has been negotiated"
	default:
		return "UNKNOWN";
	}
}

/** 
 *  @brief          Convert supplicant state to a printable string 
 *
 *  @param state    Supplicant state
 *  @return         Pointer to a printable string
 */
const char * wpa_supplicant_state_txt(int state)
{
	switch (state) {
	case WAPI_COMPLETED:
		return "0";//"Authentication Succeed;Data been encrypted."
	case WAPI_AUTHENICATING:
		return "2";//"Certificate is being authenticated...."
	case WAPI_3WAY_HANDSHAKE:
		return "7";//"Unicast Key has been negotiated"
	case WAPI_GROUP_HANDSHAKE:
		return "10";//"Multicast Key has been negotiated"
	case WAPI_DISCONNECTED:
		return "12";//"Disconnected"
	case WAPI_SCANNING:
		return "13";//"Scanning"
	case WAPI_ASSOCIATING:
		return "14";//"Associating"
	case WAPI_ASSOCIATED:
		return "15";//"Associated"
	default:
		return "UNKNOWN";
	}
}

/** 
 *  @brief          Convert wpa mode to a printable string 
 *
 *  @return         Pointer to a printable string
 */
const char * wpa_mode_txt()
{
	return "1";
}

/** 
 *  @brief          Convert wpa signal to a printable string 
 *
 *  @return         Pointer to a printable string
 */
const char * wpa_signal_txt()
{
	return "60";
}

/** 
 *  @brief          Diassociate the current connection
 *
 *  @param wpa_s    A pointer to a wapi_asue_st structure
 *  @param reason_code reason code
 *  @return            None
 */
void wpa_supplicant_disassociate(struct wapi_asue_st *wpa_s,
				 int reason_code)
{
	u8 *addr = NULL;
	wpa_s->wapi_state = WAPI_DISCONNECTED;
	if (memcmp(wpa_s->bssid, "\x00\x00\x00\x00\x00\x00", ETH_ALEN) != 0) {
		wpa_drv_disassociate(wpa_s, wpa_s->bssid, reason_code);
		addr = wpa_s->bssid;
	}
}

/** 
 *  @brief          Request association
 *
 *  @param wpa_s    A pointer to a wapi_asue_st structure
 *  @param ssid     SSID
 *  @param ssid_len Length of the SSID 
 *  @return         None
 */
void wpa_supplicant_associate(struct wapi_asue_st *wpa_s,const u8 *ssid, int ssid_len)
{
	struct wpa_driver_associate_params params;
	int i;
	struct wpa_scan_result *res;
	u8 wpa_ie[256];
	unsigned char bssid[ETH_ALEN];
     

	wpa_hexdump_ascii(MSG_DEBUG, " curent ssid:", (u8 *) ssid, ssid_len);
	memset(&params, 0, sizeof(params));
	memset(wpa_ie, 0, sizeof(wpa_ie));
	params.ssid = ssid;
	params.ssid_len = ssid_len; //strlen((char *)ssid);
	params.mode = IW_MODE_INFRA;
	
	 for (i = 0; i < wpa_s->num_scan_results; i++) {
		res = &wpa_s->scan_results[i];
		if(memcmp(params.ssid, res->ssid,res->ssid_len)==0 && res->ssid_len != 0){
			
		        wpa_s->ap_wapi_ie_len = res->wapi_ie_len;
				
		        if(res->wapi_ie_len)
		                memcpy(wpa_s->ap_wapi_ie, res->wapi_ie, res->wapi_ie_len); 
				
			wpa_printf(MSG_DEBUG, "wpa_supplicant_associate, ie len=%d\n",
                    res->wapi_ie_len);
			params.wpa_ie_len = wpa_s->assoc_wapi_ie_len;
			params.wpa_ie = wpa_s->assoc_wapi_ie;
			memcpy(bssid, res->bssid, ETH_ALEN);
			params.bssid = bssid; 
			params.freq = res->freq;
			wpa_s->wapi_state = WAPI_ASSOCIATED;
       
			if (!memcmp(wpa_s->bssid, "\x00\x00\x00\x00\x00\x00", ETH_ALEN)) /*2006-11-13*/
			{
				/* Timeout for IEEE 802.11 authentication and association */
				wpa_supplicant_req_auth_timeout(wpa_s, 10, 0);
				
				if(wpa_drv_associate(wpa_s, &params)) 
					wpa_printf(MSG_ERROR,"wpa_supplicant_associate failed\n");
				
				wpa_s->wapi_state = WAPI_DISCONNECTED;
			}
			break;
		}
	}
	if(i == wpa_s->num_scan_results) {
		wpa_printf(MSG_DEBUG, "No suitable AP found.");
		wpa_supplicant_req_scan(wpa_s, 5, 0);
	}
}

/** 
 *  @brief          Authentication timeout handler function
 *
 *  @param eloop_ctx    A pointer to a wapi_asue_st structure
 *  @param timeout_ctx  Timeout context
 *  @return         None
 */
static void wpa_supplicant_timeout(void *eloop_ctx, void *timeout_ctx)
{
	struct wapi_asue_st *wpa_s = eloop_ctx;

	
	wpa_msg(wpa_s, MSG_INFO, "Authentication with " MACSTR " timed out.",
		MAC2STR(wpa_s->bssid));
	//wpa_blacklist_add(wpa_s, wpa_s->bssid);
	wpa_supplicant_disassociate(wpa_s, REASON_DEAUTH_LEAVING);
	wpa_s->reassociate = 1;
	wpa_supplicant_req_scan(wpa_s, 5, 0);
}

/** 
 *  @brief          Setting authentication timeout
 *
 *  @param wpa_s  A pointer to a wapi_asue_st structure
 *  @param sec    Number of seconds
 *  @param usec   Number of microseconds
 *  @return       None
 */
void wpa_supplicant_req_auth_timeout(struct wapi_asue_st *wpa_s,
				     int sec, int usec)
{
	wpa_msg(wpa_s, MSG_DEBUG, "Setting authentication timeout: %d sec "
		"%d usec", sec, usec);
	eloop_cancel_timeout(wpa_supplicant_timeout, wpa_s, NULL);
	eloop_register_timeout(sec, usec, wpa_supplicant_timeout, wpa_s, NULL);
}

/** 
 *  @brief        Send Ethernet frame
 *
 *  @param wpa_s  A pointer to a wpa_supplicant data
 *  @param dest   Destination MAC address
 *  @param proto  Ethertype
 *  @param buf    Frame payload starting from IEEE 802.1X header
 *  @param len    Frame payload length
 *  @return       0 on success, otherwise -1
 */
int wpa_ether_send(struct wapi_asue_st *wpa_s, const u8 *dest, u16 proto,
		   const u8 *buf, size_t len)
{
	int res = 0;
	assert(wpa_s->l2 != NULL);

	res = l2_packet_send(wpa_s->l2, dest, proto, buf, len);
	if(  res == 0)
	{
		if(wpa_s->wapi_state < WAPI_COMPLETED)
			wpa_set_retry(wpa_s, dest, proto, buf, len);
	}
 	return res;
}

/** 
 *  @brief        Cancel authentication timeout
 *
 *  @param wpa_s  A pointer to a wpa_supplicant data
 *  @return       None
 */
void wpa_supplicant_cancel_auth_timeout(struct wapi_asue_st *wpa_s)
{
	wpa_msg(wpa_s, MSG_DEBUG, "Cancelling authentication timeout");
	eloop_cancel_timeout(wpa_supplicant_timeout, wpa_s, NULL);
	//wpa_blacklist_del(wpa_s, wpa_s->bssid);
}

/** 
 *  @brief        Handle supplicant event
 *
 *  @param wpa_s  A pointer to a wpa_supplicant data
 *  @param event  Supplicant event
 *  @param data   Event data
 *  @return       None
 */
void wpa_supplicant_event(struct wapi_asue_st *wpa_s, wpa_event_type event,
			  void *data)
{
	u8 bssid[ETH_ALEN];
	
	
	wpa_s->wai_err = WAI_ERR_INIT;
	
	wpa_drv_get_bssid(wpa_s, bssid);    

	switch (event) {
	case EVENT_ASSOC:
	{
		MAC_ADDRESS bssid_s;
		MAC_ADDRESS own_s;

		memset(&bssid_s, 0, sizeof(bssid_s));
		memset(&own_s, 0, sizeof(own_s));

		memcpy(bssid_s.v, bssid, sizeof(bssid_s.v));
		memcpy(wpa_s->bssid, bssid, sizeof(bssid));
		memcpy(own_s.v, wpa_s->own_addr, sizeof(own_s.v));
	        wpa_hexdump(MSG_DEBUG,"bssid",bssid, sizeof(bssid));
	        wpa_hexdump(MSG_DEBUG,"own mac",wpa_s->own_addr, 6);
	        if(wpa_s->ap_wapi_ie_len)
			    WAI_Msg_Input(CONN_ASSOC, &bssid_s, &own_s, wpa_s->ap_wapi_ie, wpa_s->ap_wapi_ie_len);
	        else
	            WAI_Msg_Input(CONN_ASSOC, &bssid_s, &own_s, NULL, 0);
		wpa_supplicant_cancel_auth_timeout(wpa_s);
		wpa_s->wapi_state = WAPI_ASSOCIATED;
		break;
	}
	case EVENT_DISASSOC:
		wpa_printf(MSG_DEBUG, "Disconnect event ");

		MAC_ADDRESS bssid_s;
		MAC_ADDRESS own_s;

		memset(&bssid_s, 0, sizeof(bssid_s));
		memset(&own_s, 0, sizeof(own_s));

		memcpy(bssid_s.v, bssid, sizeof(bssid_s.v));
		memcpy(wpa_s->bssid, bssid, sizeof(bssid));
		memcpy(own_s.v, wpa_s->own_addr, sizeof(own_s.v));
		WAI_Msg_Input(CONN_DISASSOC, &bssid_s, &own_s, NULL, 0);
		wpa_printf(MSG_DEBUG, "WAI_Msg_Input  success");
	        wpa_s->wapi_state = WAPI_DISCONNECTED;
	        wpa_supplicant_req_scan(wpa_s, 5, 100000);
		break;
	case EVENT_MICHAEL_MIC_FAILURE:
		break;
	case EVENT_SCAN_RESULTS:
		wpa_supplicant_get_scan_results(wpa_s);
	        if(wpa_s->wapi_state !=WAPI_ASSOCIATED)
		        wpa_supplicant_associate(wpa_s, wpa_s->conf->ssid, wpa_s->conf->ssid_len);
		break;
	case EVENT_ASSOCINFO:
		//wpa_supplicant_associnfo(wpa_s, data);
		break;
	case EVENT_INTERFACE_STATUS:
		break;
	default:
		wpa_printf(MSG_INFO, "Unknown event %d", event);
		break;
	}
}



extern struct eloop_data eloop;

/** 
 *  @brief  Get random byte
 *
 */
unsigned char WIFI_get_rand_byte()
{
	return (unsigned char)(rand() & 0xFF);
}

/** 
 *  @brief Handle deauthentication 
 *
 */
void WIFI_Action_Deauth()
{
	struct wapi_asue_st *wpa_s = (struct wapi_asue_st *)eloop.user_data;

	wpa_supplicant_disassociate(wpa_s, 15);
}

/** 
 *  @brief        Transmit packet
 *
 *  @param pbuf   A pointer to a data buffer
 *  @param length Length
 *  @return       Length of the transmit data
 */
unsigned long WIFI_TX_packet(const char* pbuf, int length)
{
	struct wapi_asue_st *wpa_s = (struct wapi_asue_st *)eloop.user_data;
	int ret;

	ret = l2_packet_send(wpa_s->l2, wpa_s->bssid, 0x88b4, (const u8*)pbuf, (unsigned  int)length);

	if (ret < 0)
	{
		return 0;
	}
	else
	{
		return (unsigned long)ret;
	}
}

/** 
 *  @brief        Set the group key
 *
 *  @param pKeyValue Key value
 *  @param keylength Key length
 *  @param key_idx   Key index
 *  @param keyIV     Key IV
 *  @return          0 on success, -1 on failure
 */
int WIFI_group_key_set(const unsigned char* pKeyValue, int keylength, int key_idx, const unsigned char* keyIV)
{
	struct wapi_asue_st *wpa_s = (struct wapi_asue_st *)eloop.user_data;
	int alg;
	u8 msk_len;
	
	wpa_hexdump(MSG_DEBUG, "WIFI_group_key_set: KEYSC", keyIV, 16);

	alg = WAPI_ALG_SMS4;
	msk_len = 32;

	if (wpa_drv_set_key(wpa_s, alg,
				   (u8 *) "\xff\xff\xff\xff\xff\xff",
				   key_idx, 1, keyIV, 16,
				   pKeyValue, keylength) < 0) {
		wpa_printf(MSG_WARNING, "WPA: Failed to set MSK to "
			   "the driver.");
		return -1;
	}
	
	return 0;
}

/** 
 *  @brief        Set the unicast key
 *
 *  @param pKeyValue Key value
 *  @param keylength Key length
 *  @param key_idx   Key index
 *  @return          0 on success, -1 on failure
 */
int WIFI_unicast_key_set(const char* pKeyValue, int keylength, int key_idx)
{
	struct wapi_asue_st *wpa_s = (struct wapi_asue_st *)eloop.user_data;

	int alg, keylen, rsclen;
	u8 *key_rsc;
	u8 null_rsc[16] = { 0, 0, 0, 0, 0, 0, 0, 0 };

	
	wpa_printf(MSG_DEBUG, "WIFI_unicast_key_set: Installing USK to the driver.");

	alg = WAPI_ALG_SMS4;
	keylen = 32;
	rsclen = 16;

	key_rsc = null_rsc;
	
	if (wpa_drv_set_key(wpa_s, alg, wpa_s->bssid, key_idx, 1, key_rsc, rsclen,
			    (u8 *) pKeyValue, keylen) < 0) {
		wpa_printf(MSG_WARNING, "WIFI_unicast_key_set: Failed to set PTK to the driver");
		return -1;
	}

	return 0;
}

/** 
 *  @brief        Init signal action
 *
 *  @param signum  Signal number
 *  @param handle  OS timer expired handler
 *  @return        None 
 */
static void init_sigaction(int signum, OS_timer_expired handle)
{
	signal(signum, handle);
}

/** 
 *  @brief        OS timer setup
 *
 *  @param deltaTimer delta timer 
 *  @param repeated   repeat
 *  @param pfunction  OS timer expired handler
 *  @param pdata      pointer to a data buffer
 *  @return        None 
 */
void* OS_timer_setup(int deltaTimer, int repeated, OS_timer_expired pfunction, const void* pdata)
{
	struct itimerval value;
	int sec = 0, usec = 0;
	int repsec = 0, repusec = 0;
	

	if (deltaTimer < 0 || repeated < 0 || pfunction == NULL)
		return NULL;
	
	init_sigaction(SIGALRM, pfunction);

	sec  = deltaTimer/(1000);
	usec = (deltaTimer%(1000)) * 1000;

	repsec  = repeated/(1000);
	repusec = (repeated%(1000)) * 1000;

	memset(&value, 0, sizeof(value));

	value.it_interval.tv_sec   = repsec;
	value.it_interval.tv_usec = repusec;

	value.it_value.tv_sec   = sec;
	value.it_value.tv_usec = usec;

	setitimer(ITIMER_REAL, &value, NULL);

	return (void *)pfunction;
}

extern void eloop_handle_alarm(int sig);

/** 
 *  @brief        OS timer clean
 *
 *  @param pTimer timer 
 *  @return        None 
 */
void OS_timer_clean(void* pTimer)
{
	struct itimerval value;
	
	memset(&value, 0, sizeof(value));
	setitimer(ITIMER_REAL, &value, NULL);

	signal(SIGALRM, eloop_handle_alarm);
}

/** 
 *  @brief       Set WAI IE
 *
 *  @param pbuf   A pointer to the IE buffer
 *  @param length Length of the IE buffer
 *  @return       0 
 */
int WIFI_WAI_IE_set(const u8* pbuf, int length)
{
	struct wapi_asue_st *wpa_s = (struct wapi_asue_st *)eloop.user_data;
	
	wpa_printf(MSG_DEBUG, " wpa_s = '%p ", wpa_s);
	wpa_printf(MSG_DEBUG, " wpa_driver_set_wpa_ie entry ");
	wpa_drv_set_wpa_ie(wpa_s,pbuf, length);
	wpa_printf(MSG_DEBUG, " wpa_driver_set_wpa_ie end ");
	wpa_s->assoc_wapi_ie_len = length;
	memcpy(wpa_s->assoc_wapi_ie, pbuf, length);
	return 0;
}
