/** @file wapi_asue_retry.c
  * @brief This file contains functions for wapi_asue_retry
 *
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>

#include "include/common.h"
#include "include/wapi.h"
#include "include/wapi_asue.h"
#include "include/eloop.h"
#include "include/l2_packet.h"
#include "include/wapi_asue_retry.h"

static int get_retry_time(struct wai_hdr *hdr);
static void wpa_asue_req_retry(struct wapi_asue_st *wpa_s, struct wapi_retry *retry, int sec, int usec);
static void wpa_asue_retry(void * eloop_ctx, void * timeout_ctx);
static void wpa_erase_retry(struct wapi_asue_st *wpa_s);
static int wpa_retry_send(struct l2_packet_data *l2, struct wapi_retry *retry);

/** 
 *  @brief           Get retry time 
 *
 *  @param hdr       A Pointer to wai_hdr structure
 *  @return          retry time in second
 */
static int get_retry_time(struct wai_hdr *hdr)
{
	int secs = 0;
	switch(hdr->type)
	{
		case WAI_ACCESS_AUTH_REQUEST:
			secs = 31;
			break;
		case WAI_USK_NEGOTIATION_RESPONSE:
		case WAI_MSK_ANNOUNCEMENT_RESPONSE:
			secs = 1;
			break;
		default:
			secs = 0;
			break;
	}
	return secs;
}

/** 
 *  @brief           Setting retry request 
 *
 *  @param wpa_s     A Pointer to wapi_asue_st structure
 *  @param retry     A Pointer to wapi_retry structure
 *  @param sec       Number of seconds
 *  @param usec      Number of microseconds
 *  @return          None
 */
static void wpa_asue_req_retry(struct wapi_asue_st *wpa_s, struct wapi_retry *retry, int sec, int usec)
{
	wpa_msg(wpa_s, MSG_DEBUG, "Setting retry request: %d sec %d usec",
		sec, usec);
	eloop_cancel_timeout(wpa_asue_retry, wpa_s, retry);
	eloop_register_timeout(sec, usec, wpa_asue_retry, wpa_s, retry);
}

/** 
 *  @brief           Setting retry 
 *
 *  @param eloop_ctx   A Pointer to wapi_asue_st structure
 *  @param timeout_ctx A Pointer to wapi_retry structure
 *  @return          None
 */
static void wpa_asue_retry(void * eloop_ctx, void * timeout_ctx)
{
	struct wapi_asue_st *wpa_s = (struct wapi_asue_st *)eloop_ctx;
	struct wapi_retry *retry = (struct wapi_retry *)timeout_ctx;

	retry->handle(wpa_s->l2, retry);
	if(retry->retry_count <3)
		wpa_asue_req_retry(wpa_s, retry, get_retry_time((struct wai_hdr *)(retry->retry_buf)), 0);
	else
	{	
		wpa_printf(MSG_DEBUG, "resend counter is bigger than 3, then deauthenation with AP");
		wpa_supplicant_deauthenticate(wpa_s, REASON_RETRY_TIMEOUT);
	}
}

/** 
 *  @brief           Erase retry 
 *
 *  @param wpa_s     A Pointer to wapi_asue_st structure
 *  @return          None
 */
static void wpa_erase_retry(struct wapi_asue_st *wpa_s)
{
	struct wapi_retry *retry = &wpa_s->wapi_sm->retry;
		
	eloop_cancel_timeout(wpa_asue_retry, wpa_s, retry);
	memset(retry, 0, sizeof(*retry));
	return;
}

/** 
 *  @brief           Send retry 
 *
 *  @param l2        A Pointer to l2_packet_data structure
 *  @param retry     A Pointer to wapi_retry structure
 *  @return          Length of the send data
 */
static int wpa_retry_send(struct l2_packet_data *l2, struct wapi_retry *retry)
{
	retry->retry_count++;
	wpa_printf(MSG_DEBUG, "Timeout. resend to("MACSTR"), send_count=%d",
		MAC2STR(retry->mac), retry->retry_count);
	return l2_packet_send(l2, retry->mac, retry->proto, retry->retry_buf, retry->len);
}

/** 
 *  @brief           Set retry 
 *
 *  @param wpa_s     A Pointer to wapi_asue_st structure
 *  @param dest      Destination address
 *  @param proto     Protocol
 *  @param buf       Buffer
 *  @param len       Length of the buf
 *  @return          None
 */
void wpa_set_retry(struct wapi_asue_st *wpa_s,const u8 *dest, u16 proto,
		   const u8 *buf, size_t len)
{
	struct wapi_state_machine *sm = wpa_s->wapi_sm;

	
	wpa_erase_retry(wpa_s);
	memcpy(sm->retry.retry_buf, buf, len);
	sm->retry.len = len;
	sm->retry.priv = (void *)wpa_s;
	sm->retry.states = wpa_s->wapi_state;
	sm->retry.proto = proto;
	sm->retry.retry_count = 1;
	sm->retry.handle = wpa_retry_send;
	wpa_asue_req_retry(wpa_s, &sm->retry, get_retry_time((struct wai_hdr *)buf), 0);
	
}

/** 
 *  @brief           Cancel retry 
 *
 *  @param wpa_s     A Pointer to wapi_asue_st structure
 *  @param stype     header type
 *  @return          None
 */
void wpa_cancel_retry(struct wapi_asue_st *wpa_s, u8 stype)
{
	struct wapi_retry *retry = &wpa_s->wapi_sm->retry;
	struct wai_hdr *hdr = (struct wai_hdr *)retry->retry_buf;

	wpa_printf(MSG_DEBUG, "WPA: in %s:%d", __func__, __LINE__);
	if(stype == hdr->stype)
		eloop_cancel_timeout(wpa_asue_retry, wpa_s, retry);
}
