/** @file wapi_config.c
  * @brief This file contains functions for wapi_config
 *
 *  Copyright (C) 2001-2008, Iwncomm Ltd.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>

#include "include/common.h"
#include "include/wapi_config.h"
#include "include/config.h"
#include "include/wapi.h"
#include "include/eloop.h"
#include "include/interface_inout.h"

extern struct eloop_data eloop;

static int change_cert_format(const  char *cert_file, 
						unsigned char *out_user, 
						int len_user,
						unsigned char *out_asu,
						int len_asu);

/** 
 *  @brief    Free wapi config  
 *
 *  @param config   A pointer to a wapi_config structure
 *  @return         None 
 */
void wpa_config_free(struct wapi_config *config)
{
	struct wapi_config *prev = NULL;
	prev = config;
	if (prev) {
		if(prev->ssid)
			free(prev->ssid);
		if(prev->cert_name)
			free(prev->cert_name);
		free(config);
	}
}

/** 
 *  @brief    Save User certificate status  
 *
 *  @param fileconfig config file
 *  @param cert_flag  Cert status
 *  @return           0 on success, otherwise failure 
 */
int save_cert_status(char *fileconfig, char *cert_flag)
{
	int res = 0;
	res = !save_global_conf(SEP_EQUAL,fileconfig,"",  "CERT_STATUS",cert_flag);
	if(res != 0)
	{
		wpa_printf(MSG_DEBUG, "open file %s error\n", fileconfig);
	}
	return res;
}

/** 
 *  @brief          Read WPA config file  
 *
 *  @param config_file config file
 *  @return            A pointer to a wapi_config 
 */
struct wapi_config * wpa_config_read(const char *config_file)
{
	char buf[256] = {0,};
	prop_data properties[KEYS_MAX];
	int prop_count=0;
	struct wapi_config *config = NULL;
	CNTAP_PARA lib_param;  
	u8 psk_type = 0;
	
	config = malloc(sizeof(*config));
	if (config == NULL)
		return NULL;
	memset(config, 0, sizeof(*config));

	wpa_printf(MSG_DEBUG, "Reading configuration file '%s'",
		   config_file);
	prop_count=load_prop(SEP_EQUAL, (char *)config_file,properties);	

	if(prop_count == 0){
		wpa_printf(MSG_ERROR, "no keys and values pair in configuration file '%s'",
			   config_file);
		return NULL;
	}

	bzero(buf, 256);
	
	/*Get Authentication and Key Management Suites*/
	get_prop("WAPI", buf, properties, prop_count);
	config->wapi_policy = atoi(buf);
	wpa_printf(MSG_DEBUG, "wapi_policy '%d'",   config->wapi_policy);
	
	if (config->wapi_policy & 0x08)
	{
		lib_param.authType = AUTH_TYPE_WAPI;
	}
	else
		if (config->wapi_policy & 0x04)
		{
			lib_param.authType = AUTH_TYPE_WAPI_PSK;
		}
		else
		{
			lib_param.authType = AUTH_TYPE_NONE_WAPI;
		}
	
	wpa_printf(MSG_DEBUG, "WAPI '%d'", config->wapi_policy);

	if(config->wapi_policy != 0)
	{
		config->group_cipher = WAPI_CIPHER_SMS4;
		config->pairwise_cipher = WAPI_CIPHER_SMS4;
	}
	bzero(buf, 256);

	get_prop("SSID", buf, properties, prop_count);
	config->ssid_len = strlen(buf);
	config->ssid = (u8*)strdup(buf);
	wpa_hexdump_ascii(MSG_DEBUG, " SSID:", config->ssid, config->ssid_len);
	
	bzero(buf, 256);
	get_prop("CERT_INDEX", buf, properties, prop_count);
	eloop.cert_info.config.used_cert = (u8)atoi(buf);/*1:x509 2:GBW */

	bzero(buf, 256);
	get_prop("CERT_NAME", eloop.cert_info.config.cert_name, properties, prop_count);
	wpa_printf(MSG_DEBUG, "CERT_NAME '%s'", eloop.cert_info.config.cert_name);

	if((config->wapi_policy & 0x08) && strlen(eloop.cert_info.config.cert_name))
	{		

		change_cert_format(eloop.cert_info.config.cert_name, 
							lib_param.para.user, 2048,
							lib_param.para.as, 2048);
	}
	
	memset(buf, 0, sizeof(buf));
	get_prop("PSK_KEY_TYPE", buf, properties, prop_count);
	psk_type = (u8)atoi(buf);
	wpa_printf(MSG_DEBUG, "PSK_KEY_TYPE '%d'",   psk_type);

	bzero(buf, 256);
	get_prop("PSK", buf, properties, prop_count);
	wpa_printf(MSG_DEBUG, "PSK'%s'", buf);
	
	if(config->wapi_policy & WAPI_KEY_MGMT_PSK){ /*bit 2:PSK*/
		
		lib_param.para.kt = psk_type;
		lib_param.para.kl = strlen(buf);
		memcpy(lib_param.para.kv, buf, lib_param.para.kl);
	}

	WAI_CNTAPPARA_SET(&lib_param);	
	
	free_prop(properties,prop_count);
	return config;
}

/** 
 *  @brief           Find mark in source  
 *
 *  @param src      Source
 *  @param lsrc     Length of the source
 *  @param mark     mark string
 *  @param lmark    length of the mark
 *  @return         position of the mark if found, otherwise NULL 
 */
static const unsigned char* findmark_mem(const unsigned char* src, int lsrc, const unsigned char* mark, int lmark)
{
	const unsigned char* p = src;
	const unsigned char* pe = src+lsrc;
	if (NULL==src || NULL==mark || lsrc<0 || lmark<0 || lsrc<lmark)
	{
		return NULL;
	}
	pe -= lmark;
	for (; p<=pe; p++)
	{
		if (0 == memcmp(p, mark, lmark))
		{
			return p;
		}
	}
	return NULL;
}

/** 
 *  @brief           Find string mark in source  
 *
 *  @param src      Source
 *  @param lsrc     Length of the source
 *  @param mark     mark string
 *  @return         position of the mark if found, otherwise NULL 
 */
static const unsigned char* findstrmark_mem(const unsigned char* src, int lsrc, const char* mark)
{
	return findmark_mem(src, lsrc, (const unsigned char*)mark, strlen(mark));
}

/** 
 *  @brief           Change certificate format  
 *
 *  @param cert_file Certificate file
 *  @param out_user  user certificate
 *  @param len_user  length of user certificate
 *  @param out_asu   ASU certificate
 *  @param len_user  Length of ASU certificate
 *  @return         0 on success, -1 on failure
 */
//changed by kaibing.chen :use two certificates rathen than one simple certifcate for WAPI feature. 
static int change_cert_format(const  char *cert_file, 
						unsigned char *out_user, 
						int len_user,
						unsigned char *out_asu,
						int len_asu)
{
	int ret = -1;
	int len = 0;
	unsigned char buf_asu[1024*8] = {0};
	unsigned char buf_user[1024*8] = {0};

	const char* mark_ec_e = "-----END EC PRIVATE KEY-----";
	const char* mark_s = "-----BEGIN CERTIFICATE-----";
	const char* mark_e = "-----END CERTIFICATE-----";

	char *asu_cert_file = NULL;
	char *user_cert_file = NULL;

	if (cert_file == NULL || out_user == NULL || out_asu == NULL || len_user <= 0 || len_asu <= 0)
	{
		wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
		return ret;
	}

	{
		FILE *asu_fp = NULL;
		FILE *user_fp = NULL;
		if( (user_cert_file = strchr( cert_file , ';' )) == NULL )
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		*user_cert_file++ = '\0';
		asu_cert_file = cert_file;
		if( asu_cert_file == NULL )
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		asu_fp = fopen((const char*)asu_cert_file, "rb");
		if (asu_fp == NULL)
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		len = fread(buf_asu, 1, 1024*8, asu_fp);
		fclose(asu_fp);
		if (len <= 0)
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		user_fp = fopen((const char*)user_cert_file, "rb");
		if (user_fp == NULL)
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		len = fread(buf_user, 1, 1024*8, user_fp);
		fclose(user_fp);
		if (len <= 0)
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
	}

	{
		unsigned char *p  = NULL;
		const unsigned char *ps  = NULL;
		const unsigned char *pe  = NULL;
		ps = findstrmark_mem(buf_asu, len, mark_s);
		pe = findstrmark_mem(buf_asu, len, mark_e);
		if (NULL==ps || NULL==pe || ps>=pe)
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		p = out_asu;
		memcpy(p, ps, pe - ps + strlen(pe));
		wpa_printf(MSG_DEBUG, "%s",   p);

		ps = findstrmark_mem(buf_user, len, mark_s);
		pe = findstrmark_mem(buf_user, len, mark_ec_e);
		if (NULL==ps || NULL==pe || ps>=pe)
		{
			wpa_printf(MSG_DEBUG, "%s,%d\n",   __FUNCTION__,__LINE__);
			return ret;
		}
		p = out_user;
		memcpy(p, ps, pe - ps +strlen( pe ));
		wpa_printf(MSG_DEBUG, "%s",   p);
	}
	ret = 0;
	return ret;
}
