/*
 * wpa_supplicant/hostapd control interface library
 * Copyright (c) 2004-2007, Jouni Malinen <j@w1.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README and COPYING for more details.
 */

#include "includes.h"

#ifdef CONFIG_CTRL_IFACE

#ifdef CONFIG_CTRL_IFACE_UNIX
#include <sys/un.h>
#endif /* CONFIG_CTRL_IFACE_UNIX */
#ifdef ANDROID
#include <dirent.h>
#include <linux/limits.h>
#include <cutils/sockets.h>
#include <cutils/memory.h>
#include "private/android_filesystem_config.h"
#endif

#include "wpa_ctrl.h"
#include "common.h"

//added by derek
#include "cutils/log.h"

#undef LOG_TAG
#define LOG_TAG "WPA_CTRL"
//#define WPA_SUPPLICANT_DEBUG

static pthread_mutex_t wpa_ctrl_mutex = PTHREAD_MUTEX_INITIALIZER;

#undef ENTER
#undef LEAVE
#ifdef WPA_SUPPLICANT_DEBUG
#define ENTER() LOGD("enter into %s", __FUNCTION__)
#define LEAVE() LOGD("leave from %s", __FUNCTION__)
#else
#define ENTER() do { } while (0)
#define LEAVE() do { } while (0)
#endif

#if defined(CONFIG_CTRL_IFACE_UNIX) || defined(CONFIG_CTRL_IFACE_UDP)
#define CTRL_IFACE_SOCKET
#ifdef ANDROID
static const char *local_socket_dir = "/data/misc/wifi/sockets/clients";
static const char *local_socket_prefix = "wpa_ctrl_";
#endif /* ANDROID */
#endif /* CONFIG_CTRL_IFACE_UNIX || CONFIG_CTRL_IFACE_UDP */

/**
 * struct wpa_ctrl - Internal structure for control interface library
 *
 * This structure is used by the wpa_supplicant/hostapd control interface
 * library to store internal data. Programs using the library should not touch
 * this data directly. They can only use the pointer to the data structure as
 * an identifier for the control interface connection and use this as one of
 * the arguments for most of the control interface library functions.
 */
struct wpa_ctrl {
#ifdef CONFIG_CTRL_IFACE_UDP
	int s;
	struct sockaddr_in local;
	struct sockaddr_in dest;
	char *cookie;
#endif /* CONFIG_CTRL_IFACE_UDP */
#ifdef CONFIG_CTRL_IFACE_UNIX
	int s;
	struct sockaddr_un local;
	struct sockaddr_un dest;
#endif /* CONFIG_CTRL_IFACE_UNIX */
#ifdef CONFIG_CTRL_IFACE_NAMED_PIPE
	HANDLE pipe;
#endif /* CONFIG_CTRL_IFACE_NAMED_PIPE */
};


#ifdef CONFIG_CTRL_IFACE_UNIX

struct wpa_ctrl * wpa_ctrl_open(const char *ctrl_path)
{
	struct wpa_ctrl *ctrl;
	static int counter = 0;

    ENTER();
	ctrl = os_malloc(sizeof(*ctrl));
	if (ctrl == NULL)
		return NULL;
	os_memset(ctrl, 0, sizeof(*ctrl));

	ctrl->s = socket(PF_UNIX, SOCK_DGRAM, 0);
	if (ctrl->s < 0) {
		os_free(ctrl);
		return NULL;
	}

	ctrl->local.sun_family = AF_UNIX;
	os_snprintf(ctrl->local.sun_path, sizeof(ctrl->local.sun_path),
#ifdef ANDROID	
		    "%s/%s%d-%d", local_socket_dir, local_socket_prefix,
                    getpid(), counter++);
#else /* ANDROID */
		    "/tmp/wpa_ctrl_%d-%d", getpid(), counter++);
#endif		    
    LOGD("in wpa_ctrl_open, ctrl->local.sun_path is %s", ctrl->local.sun_path);

    //modified by derek
	if (bind(ctrl->s, (struct sockaddr *) &ctrl->local, sizeof(ctrl->local)) < 0) {
		LOGD("in wpa_ctrl_open, bind(PF_UNIX) fails!");

		if (connect(ctrl->s, (struct sockaddr *) &ctrl->local, sizeof(ctrl->local)) < 0) {
			LOGD("in wpa_ctrl_open, local socket file exists, but does not"
				   " allow connections - assuming it was left"
				   "over from forced program termination");
			if (unlink(ctrl->local.sun_path) < 0) {
				LOGD("in wpa_ctrl_open, could not unlink "
					   "existing '%s'",
					   ctrl->local.sun_path);
		        close(ctrl->s);
		        os_free(ctrl);
				return NULL;
			} else {
                LOGD("in wpa_ctrl_open, unlink successfully");
            }

			if (bind(ctrl->s, (struct sockaddr *) &ctrl->local, sizeof(ctrl->local)) < 0) {
                LOGD("in wpa_ctrl_open, after unlink successfully, bind fails!");
		        close(ctrl->s);
		        os_free(ctrl);
                return NULL;
			} else {
                LOGD("in wpa_ctrl_open, after unlink successfully, bind successfully");
            }
		} else {
			LOGD("in wpa_ctrl_open, this socket file exists and seems to "
				   "be in use - cannot override it");
			LOGD("in wpa_ctrl_open, delete '%s' manually if it is "
				   "not used anymore", ctrl->local.sun_path);
		    close(ctrl->s);
		    os_free(ctrl);
            return NULL;
		}
	} else {
		LOGD("in wpa_ctrl_open, bind(PF_UNIX) successfully!");
    }

#ifdef ANDROID
        chmod(ctrl->local.sun_path, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
        chown(ctrl->local.sun_path, AID_SYSTEM, AID_WIFI);
	/*
	 * If the ctrl_path isn't an absolute pathname, assume that
	 * it's the name of a socket in the Android reserved namespace.
	 * Otherwise, it's a normal UNIX domain socket appearing in the
	 * filesystem.
	 */
	if (ctrl_path != NULL && *ctrl_path != '/') {
		os_snprintf(ctrl->dest.sun_path, sizeof(ctrl->dest.sun_path), "wpa_%s",
			    ctrl_path);
		if (socket_local_client_connect(ctrl->s,
						ctrl->dest.sun_path,
						ANDROID_SOCKET_NAMESPACE_RESERVED,
						SOCK_DGRAM) < 0) {
            LOGD("in wpa_ctrl_open, socket_local_client_connect fails!");
			close(ctrl->s);
			unlink(ctrl->local.sun_path);
			os_free(ctrl);
			return NULL;
		}
        else
        {
            LOGD("in wpa_ctrl_open, socket_local_client_connect successfully!");
        }
		return ctrl;
	}
#endif
	ctrl->dest.sun_family = AF_UNIX;
	os_snprintf(ctrl->dest.sun_path, sizeof(ctrl->dest.sun_path), "%s",
		    ctrl_path);
    LOGD("in wpa_ctrl_open, ctrl->dest.sun_path is %s", ctrl->dest.sun_path);

	if (connect(ctrl->s, (struct sockaddr *) &ctrl->dest,
		    sizeof(ctrl->dest)) < 0) {
        LOGD("in wpa_ctrl_open, connect fails!");
		close(ctrl->s);
		unlink(ctrl->local.sun_path);
		os_free(ctrl);
		return NULL;
	} else {
        LOGD("in wpa_ctrl_open, connect successfully");
    }
    LEAVE();

	return ctrl;
}


void wpa_ctrl_close(struct wpa_ctrl *ctrl)
{
    ENTER();
	unlink(ctrl->local.sun_path);
	close(ctrl->s);
	os_free(ctrl);
    LEAVE();
}

#ifdef ANDROID
/**
 * wpa_ctrl_cleanup() - Delete any local UNIX domain socket files that
 * may be left over from clients that were previously connected to
 * wpa_supplicant. This keeps these files from being orphaned in the
 * event of crashes that prevented them from being removed as part
 * of the normal orderly shutdown.
 */
void wpa_ctrl_cleanup(void)
{
    DIR *dir;
    struct dirent entry;
    struct dirent *result;
    size_t dirnamelen;
    int prefixlen = strlen(local_socket_prefix);
    size_t maxcopy;
    char pathname[PATH_MAX];
    char *namep;

    ENTER();
    if ((dir = opendir(local_socket_dir)) == NULL)
        return;

    dirnamelen = (size_t)snprintf(pathname, sizeof(pathname), "%s/", local_socket_dir);
    if (dirnamelen >= sizeof(pathname)) {
        closedir(dir);
        return;
    }
    namep = pathname + dirnamelen;
    maxcopy = PATH_MAX-dirnamelen;
    while (readdir_r(dir, &entry, &result) == 0 && result != NULL) {
        if (strncmp(entry.d_name, local_socket_prefix, prefixlen) == 0) {
            if (strlcpy(namep, entry.d_name, maxcopy) < maxcopy) {
                unlink(pathname);
            }
        }
    }
    closedir(dir);
    LEAVE();
}
#endif /* ANDROID */

#else /* CONFIG_CTRL_IFACE_UNIX */
#ifdef ANDROID
void wpa_ctrl_cleanup()
{
}
#endif /* ANDROID */
#endif /* CONFIG_CTRL_IFACE_UNIX */

#ifdef CONFIG_CTRL_IFACE_UDP

struct wpa_ctrl * wpa_ctrl_open(const char *ctrl_path)
{
	struct wpa_ctrl *ctrl;
	char buf[128];
	size_t len;

    ENTER();
	ctrl = os_malloc(sizeof(*ctrl));
	if (ctrl == NULL)
		return NULL;
	os_memset(ctrl, 0, sizeof(*ctrl));

	ctrl->s = socket(PF_INET, SOCK_DGRAM, 0);
	if (ctrl->s < 0) {
		perror("socket");
		os_free(ctrl);
		return NULL;
	}

	ctrl->local.sin_family = AF_INET;
	ctrl->local.sin_addr.s_addr = htonl((127 << 24) | 1);
	if (bind(ctrl->s, (struct sockaddr *) &ctrl->local,
		 sizeof(ctrl->local)) < 0) {
		close(ctrl->s);
		os_free(ctrl);
		return NULL;
	}

	ctrl->dest.sin_family = AF_INET;
	ctrl->dest.sin_addr.s_addr = htonl((127 << 24) | 1);
	ctrl->dest.sin_port = htons(WPA_CTRL_IFACE_PORT);
	if (connect(ctrl->s, (struct sockaddr *) &ctrl->dest,
		    sizeof(ctrl->dest)) < 0) {
		perror("connect");
		close(ctrl->s);
		os_free(ctrl);
		return NULL;
	}

	len = sizeof(buf) - 1;
	if (wpa_ctrl_request(ctrl, "GET_COOKIE", 10, buf, &len, NULL) == 0) {
		buf[len] = '\0';
		ctrl->cookie = strdup(buf);
	}
    LEAVE();

	return ctrl;
}


void wpa_ctrl_close(struct wpa_ctrl *ctrl)
{
    ENTER();
	close(ctrl->s);
	os_free(ctrl->cookie);
	os_free(ctrl);
    LEAVE();
}

#endif /* CONFIG_CTRL_IFACE_UDP */


#ifdef CTRL_IFACE_SOCKET
static int get_cmd_timeout(const char * cmd, size_t  cmd_len)
{
	const int default_cmd_timeout = 4;
	int id, cmdtemp_len;
	char *name, *cmdtemp;
	int ret = default_cmd_timeout;

	if( NULL == cmd )
		return ret;

	if( os_strncmp(cmd, "GET_NETWORK ", 12) == 0) 
	{
		cmdtemp_len = cmd_len-12;
		cmdtemp = (char *)malloc( ( cmdtemp_len + 1 )* sizeof( char ) );
		memset( cmdtemp, 0, ( cmdtemp_len + 1 ) * sizeof( char  ) );
		memcpy( cmdtemp, cmd+12, cmdtemp_len );
		
		/* cmd: "<network id> <variable name>" */
		name = os_strchr(cmdtemp, ' ');
		if (name == NULL )
		{
			free( cmdtemp );
			return ret; 
		}
		*name++ = '\0';

		id = atoi(cmdtemp);
		
		LOGD( "get_cmd_timeout: GET_NETWORK id=%d name='%s'",id, name);

	        if( 0 == os_strcmp( name, "all" ) ) {
			ret = 8;
		}
		free( cmdtemp );
	}
	return ret;
	
}
int wpa_ctrl_request(struct wpa_ctrl *ctrl, const char *cmd, size_t cmd_len,
		     char *reply, size_t *reply_len,
		     void (*msg_cb)(char *msg, size_t len))
{
	struct timeval tv;
	int res;
	fd_set rfds;
	const char *_cmd;
	char *cmd_buf = NULL;
	size_t _cmd_len;
	int cmd_timeout;
    int i = 0;
    static u_short index = 0;

    ENTER();
    pthread_mutex_lock(&wpa_ctrl_mutex);

#ifdef CONFIG_CTRL_IFACE_UDP
	if (ctrl->cookie) {
		char *pos;
		_cmd_len = strlen(ctrl->cookie) + 1 + cmd_len;
		cmd_buf = os_malloc(_cmd_len );
		if (cmd_buf == NULL) {
		    pthread_mutex_unlock(&wpa_ctrl_mutex);
			return -1;
		}
		_cmd = cmd_buf;
		pos = cmd_buf;
		strcpy(pos, ctrl->cookie);
		pos += strlen(ctrl->cookie);
		*pos++ = ' ';
		memcpy(pos, cmd, cmd_len);
	} else
#endif /* CONFIG_CTRL_IFACE_UDP */

	//add index to each message to avoid mismatch of message and reply by derek.
	//each index size is 2 bytes. the message format is as belows.
	//  | index |      cmd      |
	//  |<--2-->|<-- cmd_len -->|
#if 0
	{
		_cmd = cmd;
		_cmd_len = cmd_len;
	}
#else
	{
		char *pos;
		index ++;
		if (index >= 0xffff) {
			LOGD("in wpa_ctrl_request, index is reset to 1!");
			index = 1;
		}

		//first byte should not be 0x3c
		if (*(char *)&index == '<') {
			LOGD("in wpa_ctrl_request, index skips one for < symbol!");
			index ++;
		}
		_cmd_len = cmd_len + 2;
		cmd_buf = os_malloc(_cmd_len + 1);
		if (cmd_buf == NULL) {
		    pthread_mutex_unlock(&wpa_ctrl_mutex);
			return -1;
		}
		cmd_buf[_cmd_len] = '\0';
		_cmd = cmd_buf;
		pos = cmd_buf;
		memcpy(pos, &index, 2);
		pos += 2;
		memcpy(pos, cmd, cmd_len);
	}
#endif

	cmd_timeout = get_cmd_timeout( cmd, cmd_len );

	LOGV("the command is %s,timeout is %d. ", _cmd + 2,cmd_timeout);

	if (send(ctrl->s, _cmd, _cmd_len, 0) < 0) {
		LOGD("in wpa_ctrl_request, send fails, will return -3 to caller!");
		perror("send");
		os_free(cmd_buf);
		pthread_mutex_unlock(&wpa_ctrl_mutex);
		return -3;
	}
	os_free(cmd_buf);

	
	for (;;) {
//#ifdef ANDROID
//		tv.tv_sec = 10;
//#else
		tv.tv_sec = cmd_timeout;
//#endif
		tv.tv_usec = 0;
		FD_ZERO(&rfds);
		FD_SET(ctrl->s, &rfds);
		res = select(ctrl->s + 1, &rfds, NULL, NULL, &tv);
		if (FD_ISSET(ctrl->s, &rfds)) {
			res = recv(ctrl->s, reply, *reply_len, 0);
            //LOGD("in wpa_ctrl_request, recv len is %d", res);
            if(res > 0 && reply[0] != '<')
            {
                u_short res_index = 0;
                reply[res] = '\0';
                if ( res < 2 ) {
            		LOGD("in wpa_ctrl_request, res is less than 2, it is %d", res);
			        pthread_mutex_unlock(&wpa_ctrl_mutex);
			        return -1;
                }
                //added by derek. check the index in reply is right or not.
                memcpy(&res_index, reply, 2);
            	//LOGD("in wpa_ctrl_request, index is %d, res_index is %d", index, res_index);
                if (index > res_index)
                {
            		LOGD("!!!!!!!!!!in wpa_ctrl_request, it seems to receive a previous reply,\n"
            		"!!!!!!!!!!index is %d, res_index is %d, discard this reply and\n"
            		"!!!!!!!!!!try to receive a new reply again", index, res_index);
            		continue;
                }
                else if (index < res_index) {
                	LOGD("in wpa_ctrl_request, index is less than res_index\n"
            		"index is %d, res_index is %d, try to receive again", index, res_index);
            		continue;
                }
                *reply_len = res - 2;
                memcpy(reply, reply+2, *reply_len);
                reply[*reply_len] = '\0';

                LOGV("RX buf in string is %s", reply);
            }

			if (res < 0)
            { 
				LOGD("in wpa_ctrl_request, recv fails!");
				perror("recv");
				pthread_mutex_unlock(&wpa_ctrl_mutex);
				return res;
            }
			if (res > 0 && reply[0] == '<') {
                LOGD("in wpa_ctrl_request, unsolicited message in string is %s", reply);
				/* This is an unsolicited message from
				 * wpa_supplicant, not the reply to the
				 * request. Use msg_cb to report this to the
				 * caller. */
				if (msg_cb) {
					/* Make sure the message is nul
					 * terminated. */
					if ((size_t) res == *reply_len)
						res = (*reply_len) - 1;
					reply[res] = '\0';
					msg_cb(reply, res);
				}
				continue;
			}
			*reply_len = res - 2;
			break;
		} else {
			pthread_mutex_unlock(&wpa_ctrl_mutex);
			return -2;
		}
	}
    pthread_mutex_unlock(&wpa_ctrl_mutex);
    LEAVE();

	return 0;
}
#endif /* CTRL_IFACE_SOCKET */


static int wpa_ctrl_attach_helper(struct wpa_ctrl *ctrl, int attach)
{
	char buf[10];
	int ret;
	size_t len = 10;

    ENTER();
	ret = wpa_ctrl_request(ctrl, attach ? "ATTACH" : "DETACH", 6,
			       buf, &len, NULL);
	if (ret < 0)
    {
        LOGD("in wpa_ctrl_attach_helper, the ret of wpa_ctrl_request is %d", ret);
		return ret;
    }
	if (len == 3 && os_memcmp(buf, "OK\n", 3) == 0)
		return 0;
    LEAVE();

	return -1;
}


int wpa_ctrl_attach(struct wpa_ctrl *ctrl)
{
	return wpa_ctrl_attach_helper(ctrl, 1);
}


int wpa_ctrl_detach(struct wpa_ctrl *ctrl)
{
	return wpa_ctrl_attach_helper(ctrl, 0);
}


#ifdef CTRL_IFACE_SOCKET

int wpa_ctrl_recv(struct wpa_ctrl *ctrl, char *reply, size_t *reply_len)
{
	int res;

    ENTER();
	res = recv(ctrl->s, reply, *reply_len, 0);
	if (res < 0)
		return res;
	*reply_len = res;
    LEAVE();

	return 0;
}


int wpa_ctrl_pending(struct wpa_ctrl *ctrl)
{
	struct timeval tv;
	fd_set rfds;
    ENTER();
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	FD_ZERO(&rfds);
	FD_SET(ctrl->s, &rfds);
	select(ctrl->s + 1, &rfds, NULL, NULL, &tv);
    LEAVE();

	return FD_ISSET(ctrl->s, &rfds);
}

int wpa_ctrl_pending_timeout(struct wpa_ctrl *ctrl, long sec, long usec)
{
	struct timeval tv;
	fd_set rfds;
	int n = 0;

	if (ctrl == NULL) {
		//sleep some time to make this function be pending before it returns.
		sleep(sec);
		return -1;
	}
	tv.tv_sec = sec;
	tv.tv_usec = usec;
	FD_ZERO(&rfds);
	FD_SET(ctrl->s, &rfds);
	n = select(ctrl->s + 1, &rfds, NULL, NULL, &tv);
	if (n > 0) {
		//LOGD("in wpa_ctrl_pending_timeout, n is %d", n);
		return FD_ISSET(ctrl->s, &rfds);
	}

	return n;
}

int wpa_ctrl_get_fd(struct wpa_ctrl *ctrl)
{
	return ctrl->s;
}

#endif /* CTRL_IFACE_SOCKET */


#ifdef CONFIG_CTRL_IFACE_NAMED_PIPE

#ifndef WPA_SUPPLICANT_NAMED_PIPE
#define WPA_SUPPLICANT_NAMED_PIPE "WpaSupplicant"
#endif
#define NAMED_PIPE_PREFIX TEXT("\\\\.\\pipe\\") TEXT(WPA_SUPPLICANT_NAMED_PIPE)

struct wpa_ctrl * wpa_ctrl_open(const char *ctrl_path)
{
	struct wpa_ctrl *ctrl;
	DWORD mode;
	TCHAR name[256];
	int i;

    ENTER();
	ctrl = os_malloc(sizeof(*ctrl));
	if (ctrl == NULL)
		return NULL;
	os_memset(ctrl, 0, sizeof(*ctrl));

#ifdef UNICODE
	if (ctrl_path == NULL)
		_snwprintf(name, 256, NAMED_PIPE_PREFIX);
	else
		_snwprintf(name, 256, NAMED_PIPE_PREFIX TEXT("-%S"),
			   ctrl_path);
#else /* UNICODE */
	if (ctrl_path == NULL)
		os_snprintf(name, 256, NAMED_PIPE_PREFIX);
	else
		os_snprintf(name, 256, NAMED_PIPE_PREFIX "-%s",
			    ctrl_path);
#endif /* UNICODE */

	for (i = 0; i < 10; i++) {
		ctrl->pipe = CreateFile(name, GENERIC_READ | GENERIC_WRITE, 0,
					NULL, OPEN_EXISTING, 0, NULL);
		/*
		 * Current named pipe server side in wpa_supplicant is
		 * re-opening the pipe for new clients only after the previous
		 * one is taken into use. This leaves a small window for race
		 * conditions when two connections are being opened at almost
		 * the same time. Retry if that was the case.
		 */
		if (ctrl->pipe != INVALID_HANDLE_VALUE ||
		    GetLastError() != ERROR_PIPE_BUSY)
			break;
		WaitNamedPipe(name, 1000);
	}
	if (ctrl->pipe == INVALID_HANDLE_VALUE) {
		os_free(ctrl);
		return NULL;
	}

	mode = PIPE_READMODE_MESSAGE;
	if (!SetNamedPipeHandleState(ctrl->pipe, &mode, NULL, NULL)) {
		CloseHandle(ctrl->pipe);
		os_free(ctrl);
		return NULL;
	}
    LEAVE();

	return ctrl;
}


void wpa_ctrl_close(struct wpa_ctrl *ctrl)
{
    ENTER();
	CloseHandle(ctrl->pipe);
	os_free(ctrl);
    LEAVE();
}


int wpa_ctrl_request(struct wpa_ctrl *ctrl, const char *cmd, size_t cmd_len,
		     char *reply, size_t *reply_len,
		     void (*msg_cb)(char *msg, size_t len))
{
	DWORD written;
	DWORD readlen = *reply_len;

    ENTER();
	if (!WriteFile(ctrl->pipe, cmd, cmd_len, &written, NULL))
		return -1;

	if (!ReadFile(ctrl->pipe, reply, *reply_len, &readlen, NULL))
		return -1;
	*reply_len = readlen;
    LEAVE();

	return 0;
}


int wpa_ctrl_recv(struct wpa_ctrl *ctrl, char *reply, size_t *reply_len)
{
    ENTER();
	DWORD len = *reply_len;
	if (!ReadFile(ctrl->pipe, reply, *reply_len, &len, NULL))
		return -1;
	*reply_len = len;
    LEAVE();
    
	return 0;
}


int wpa_ctrl_pending(struct wpa_ctrl *ctrl)
{
	DWORD left;

	if (!PeekNamedPipe(ctrl->pipe, NULL, 0, NULL, &left, NULL))
		return -1;
	return left ? 1 : 0;
}


int wpa_ctrl_get_fd(struct wpa_ctrl *ctrl)
{
        return ctrl->pipe;
}

#endif /* CONFIG_CTRL_IFACE_NAMED_PIPE */

#endif /* CONFIG_CTRL_IFACE */
